(*                                                                      *)
(* MLJVM - a Java Virtual Machine implemented in Standard ML            *)
(*                                                                      *)
(* http://homepages.inf.ed.ac.uk/stg/research/projects/MLJVM/           *)
(*                                                                      *)
(* Copyright (C) 2003 Stephen Gilmore and Laura Korte, The              *)
(* University of Edinburgh.                                             *)
(*                                                                      *)
(* This program is free software; you can redistribute it and/or        *)
(* modify it under the terms of the GNU General Public License          *)
(* as published by the Free Software Foundation; either version 2       *)
(* of the License, or (at your option) any later version.               *)
(*                                                                      *)
(* This program is distributed in the hope that it will be useful,      *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of       *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *)
(* GNU General Public License for more details.                         *)
(*                                                                      *)
(* You should have received a copy of the GNU General Public License    *)
(* along with this program; if not, write to the Free Software          *)
(* Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *)
(* MA 02111-1307, USA.                                                  *)
(*                                                                      *)

structure Execute :> Execute = 
struct

open Environment RuntimeClass RuntimePool

(*********************************)
(***** EXCEPTION DEFINITIONS *****)
(*********************************)

local
   val file = "Execute.sml"
in
   fun error s = Error.error file s
   fun match_error s = Error.match_error file s
   fun unimplemented_error s = Error.unimplemented_error file s 
   fun find_error s = Error.find_error file s
end


(**********************)
(***** DEBUG INFO *****)
(**********************)

val no_debugging = ref true

fun do_debug n ss = if !no_debugging
                    then case n of
		              10 => ((print ss);(TextIO.flushOut TextIO.stdOut))
			    | _  => ()
		    else case n of
		              0 => ()
			    | 1 => ()
			    | 2 => ()
			    | _ => ((print ss);(TextIO.flushOut TextIO.stdOut))


(******************************)
(***** PUSH/POP FUNCTIONS *****)
(******************************)

val push = op::
val pop  = hd

fun pushF item (FRAME frame_rec) =
    FRAME { stack          = push (item,(#stack frame_rec))
	  , local_vars     = #local_vars frame_rec
	  , current_class  = #current_class frame_rec
	  , current_method = #current_method frame_rec
	  , current_code   = #current_code frame_rec
	  , current_pc     = #current_pc frame_rec }

fun popF (FRAME frame_rec) =
    let val stack = #stack frame_rec
    in  ( FRAME { stack          = tl stack
	        , local_vars     = #local_vars frame_rec
	        , current_class  = #current_class frame_rec
	        , current_method = #current_method frame_rec
	        , current_code   = #current_code frame_rec
	        , current_pc     = #current_pc frame_rec }
	, pop stack )
    end

fun popCurrentFrame (ENV env_rec) =
    ENV { method_area    = #method_area env_rec
        , method_counter = #method_counter env_rec
	, heap           = #heap env_rec
	, stack          = tl (#stack env_rec)
	, pc_register    = #pc_register env_rec } 

fun dup (FRAME frame_rec) =
    let val stack = #stack frame_rec
    in  FRAME { stack          = (hd stack) :: stack
	      , local_vars     = #local_vars frame_rec
	      , current_class  = #current_class frame_rec
	      , current_method = #current_method frame_rec
	      , current_code   = #current_code frame_rec
	      , current_pc     = #current_pc frame_rec }   
    end


(*************************************)
(***** FUNCTIONS ON ENVIRONMENTS *****)
(*************************************)

fun getMethodArea (ENV env_rec) = #method_area env_rec    
fun getMethodCounter (ENV env_rec) = #method_counter env_rec    
fun getHeap (ENV env_rec) = #heap env_rec
fun getEnvStack (ENV env_rec) = #stack env_rec    
fun getEnvPC (ENV env_rec) = #pc_register env_rec 

fun updateMethodArea (ENV env_rec) new_ma =     
    ENV { method_area    = new_ma
        , method_counter = #method_counter env_rec
	, heap           = #heap env_rec
	, stack          = #stack env_rec
	, pc_register    = #pc_register env_rec }
fun updateHeap (ENV env_rec) heap = 
    ENV { method_area    = #method_area env_rec
        , method_counter = #method_counter env_rec
	, heap           = heap
	, stack          = #stack env_rec
	, pc_register    = #pc_register env_rec } 


(***************************)
(***** OTHER FUNCTIONS *****)
(***************************)

fun icmp_gt (i1,i2) = Int32.compare (i1,i2) = GREATER
fun icmp_ge (i1,i2) = (Int32.compare (i1,i2) = GREATER) orelse (Int32.compare (i1,i2) = EQUAL)
fun icmp_eq (i1,i2) = Int32.compare (i1,i2) = EQUAL
fun icmp_ne (i1,i2) = Int32.compare (i1,i2) <> EQUAL
fun icmp_lt (i1,i2) = Int32.compare (i1,i2) = LESS
fun icmp_le (i1,i2) = (Int32.compare (i1,i2) = LESS) orelse (Int32.compare (i1,i2) = EQUAL)

fun isNonStatic Modifiers.ACCstatic = false
  | isNonStatic _                   = true
fun isStatic Modifiers.ACCstatic = true
  | isStatic _                   = false

fun ex_try (ENV env_rec) class_name method_name =
    let val method_area = #method_area env_rec
        val class_index = Polyhash.find (#hash method_area) class_name
	val class = (valOf o Array.sub) (#array method_area,class_index)
	val methods = #methods class
	val result = SOME (Polyhash.find (#hash methods) method_name) handle _ => NONE
	fun toBool NONE     = false
	  | toBool (SOME _) = true
    in  toBool result
    end

local
   fun intToBool 0 = true
     | intToBool _ = false
in 
   fun try (hash,key) = intToBool (Polyhash.find hash key handle _ => 0)
end

fun option2bool (SOME x) = true
  | option2bool NONE     = false

fun optional_function f NONE     = NONE
  | optional_function f (SOME s) = SOME (f s)

val cp2rcp = (makeIndex o Constpool.indexValue)



(*******************************)
(***** FUNCTIONS ON FRAMES *****)
(*******************************)

fun getStack (FRAME frame_rec) = #stack frame_rec
fun getLocals (FRAME frame_rec) = #local_vars frame_rec
fun getFrameClassIdx (FRAME frame_rec) = #current_class frame_rec
(*fun getCurrentMethod (FRAME frame_rec) = #current_method frame_rec*)
fun getCode (FRAME frame_rec) = #current_code frame_rec
fun getPC (FRAME frame_rec) = #current_pc frame_rec
(*fun getOldPC (FRAME frame_rec) = #old_pc frame_rec*)

fun updateLocals (FRAME frame_rec) locals =
    FRAME { stack          = #stack frame_rec
	  , local_vars     = locals
	  , current_class  = #current_class frame_rec
	  , current_method = #current_method frame_rec
	  , current_code   = #current_code frame_rec
	  , current_pc     = #current_pc frame_rec }
fun updateCurrentPC (FRAME frame_rec) n = 
    FRAME { stack          = #stack frame_rec
	  , local_vars     = #local_vars frame_rec
	  , current_class  = #current_class frame_rec
	  , current_method = #current_method frame_rec
	  , current_code   = #current_code frame_rec
	  , current_pc     = add (#current_pc frame_rec, n) }


fun getPool m_area (FRAME frame_rec) = findCPByIndex m_area (#current_class frame_rec)



(*********************************************************)
(***** FUNCTIONS FOR EXTRACTING/UPDATING INFORMATION *****)
(*********************************************************)

fun getCurrentFrame (ENV env_rec) = hd (#stack env_rec)
fun updateCurrentFrame (ENV env_rec) frame =
    ENV { method_area    = #method_area env_rec
        , method_counter = #method_counter env_rec
	, heap           = #heap env_rec
	, stack          = push (frame, tl (#stack env_rec))
	, pc_register    = #pc_register env_rec }

fun getCurrentConstpool env =
    RuntimeClass.findCPByIndex (getMethodArea env)
                               ((getFrameClassIdx o getCurrentFrame) env)
fun updateCurrentConstpool env (idx,item) =
    let val frame = getCurrentFrame env
        val class_idx = getFrameClassIdx frame
	val ma = getMethodArea env
	val pool = RuntimeClass.findCPByIndex ma class_idx
	val _ = Array.update (pool,idx,item)
	val new_ma = updateCPByIndex ma class_idx pool
    in  updateMethodArea env new_ma
    end

fun getConstpool (ENV env_rec) classname =
    RuntimeClass.findCPByLabel (#method_area env_rec)
                               (LBL classname)

fun updateConstpool env (idx,item) class_idx =
    let val ma = getMethodArea env
	val pool = RuntimeClass.findCPByIndex ma class_idx
	val _ = Array.update (pool,idx,item)
	val new_ma = updateCPByIndex ma class_idx pool
    in  updateMethodArea env new_ma
    end
    
fun incrCurrentPC env n =
let val frame = getCurrentFrame env
in  updateCurrentFrame env (updateCurrentPC frame n)
end

fun updateFrameAndPC env frame n = updateCurrentFrame env (updateCurrentPC frame n)

fun getCurrentFields env =
    let val m_area = getMethodArea env
        val (FRAME frame_rec) = getCurrentFrame env
    in  (#fields o (findClassByIndex m_area) o #current_class) frame_rec
    end
fun getCurrentMethods env =
    let val m_area = getMethodArea env
        val (FRAME frame_rec) = getCurrentFrame env
    in  (#methods o (findClassByIndex m_area) o #current_class) frame_rec
    end

fun getClassIdx env pool idx =
    case (lookup pool idx) of
         RCP (RCPclassref {classIdx,index}) => RuntimeClass.IDX classIdx
       | CP (Constpool.CPclass idx) => ( case (lookup pool (cp2rcp idx)) of
                                              CP (Constpool.CPutf8 s) =>
					         (label2index (getMethodArea env) (LBL s)
						    handle _ => raise (Error.NoSuchClassError s))
					    | _ => raise (match_error "getClassIdx-1") )
       | _  => raise (match_error "getClassIdx-2")


(****************************************)
(***** FUNCTIONS FOR TRANSFORMATION *****)
(****************************************)

fun arg2item (arg,(items,heap)) =
    let val item    = (Heap.HPTR o Heap.getRIndex) heap
	val newHeap = Heap.insert (heap,Heap.String arg,"Ljava/lang/String;")
    in  (item :: items, newHeap)
    end
    
(* SreturnAddress = missing *)
fun stackItem2value (Sint i) = Vint i
  | stackItem2value (Sfloat f) = Vfloat f
  | stackItem2value (Sempty  ) = Vempty
  | stackItem2value (Slong l) = Vlong l 
  | stackItem2value (Sdouble d) = Vdouble d
  | stackItem2value (Sreference (Heap.HPTR h)) = Vobj h
  | stackItem2value _ = raise (unimplemented_error "stackItem2value")

fun value2stackItem (Vint    i) = Sint i
  | value2stackItem (Vempty   ) = Sempty
  | value2stackItem (Vlong   l) = Slong l
  | value2stackItem (Vfloat  f) = Sfloat f
  | value2stackItem (Vdouble d) = Sdouble d                            
  | value2stackItem (Vobj  ptr) = Sreference (Heap.HPTR ptr)

fun idx2stackItem idx env =
let val pool = getCurrentConstpool env 
in  case (lookup pool idx) of
         CP  (Constpool.CPstring s) => (Sreference o Heap.HPTR o Heap.getRIndex o getHeap) env
       | CP  (Constpool.CPint    i) => Sint i
       | CP  (Constpool.CPfloat  f) => (Sfloat o valOf o Real32.fromBytes) f
       | RCP (RCPclassref        r) => (Sreference o Heap.HPTR o Heap.getRIndex o getHeap) env
       | CP (Constpool.CPclass   c) => (Sreference o Heap.HPTR o Heap.getRIndex o getHeap) env
       | _                          => raise (match_error "idx2stackitem (LDC?)")
end



(*****************************************)
(***** FUNCTIONS FOR NAME EXTRACTION *****)
(*****************************************)

fun extractClassNameFromPool pool idx = 
    case (lookup pool idx) of
	 RCP (RCPclassref {index,classIdx}) => extract_name pool index
       | CP  (Constpool.CPclass          c) => extract_name pool (cp2rcp c)
       | _                                  => raise (match_error "extractClassNameFromPool")

fun extractNameFromNameTypeIdx pool nt_idx =
    let val (name_idx,type_idx) =
            case (lookup pool nt_idx) of
                 CP (Constpool.CPnametype {name, desc}) => (name,desc) 
	       | _                                      => raise (match_error "extractNameFromNameType-1")
	val name = case (lookup pool (cp2rcp name_idx)) of
	                CP (Constpool.CPutf8 s) => s
	              | _                       => raise (match_error "extractNameFromNameType-2")
	val desc = case (lookup pool (cp2rcp type_idx)) of
	                CP (Constpool.CPutf8 s) => s
	              | _                       => raise (match_error "extractNameFromNameType-2")
    in  name ^ desc
    end
fun extractNameFromNameTypeIdx' pool nt_idx =
    let val (name_idx,type_idx) =
            case (lookup pool nt_idx) of
                 CP (Constpool.CPnametype {name, desc}) => (name,desc) 
	       | _                                      => raise (match_error "extractNameFromNameType-1")
	val name = case (lookup pool (cp2rcp name_idx)) of
	                CP (Constpool.CPutf8 s) => s
	              | _                       => raise (match_error "extractNameFromNameType-2")
	val desc = case (lookup pool (cp2rcp type_idx)) of
	                CP (Constpool.CPutf8 s) => s
	              | _                       => raise (match_error "extractNameFromNameType-2")
    in  name
    end

fun extractCurrentClassName env =
    let val current_frame = getCurrentFrame env
        val idx = getFrameClassIdx current_frame
	val method_array = (#array o getMethodArea) env
	val current_class = valOf (Array.sub (method_array,RuntimeClass.indexValue idx))
	val pool = #pool current_class
    in  extractClassNameFromPool pool (#this current_class)
    end

fun extractCurrentClassIdx env =
    let val current_frame = getCurrentFrame env
        val idx = getFrameClassIdx current_frame
	val method_array = (#array o getMethodArea) env
	val current_class = valOf (Array.sub (method_array,RuntimeClass.indexValue idx))
	val pool = #pool current_class
    in  #this current_class
    end


(**********************************************)
(***** FUNCTIONS FOR FIELD INITIALISATION *****)
(**********************************************) 

fun init_field' pool isInstance (field:member) =
    let val desc = lookup pool (#desc field)
        val value = case desc of
	                 CP (Constpool.CPutf8 "I") => Vint (Int32.fromInt 0)
		       | CP (Constpool.CPutf8 "L") => Vlong (Int64.fromInt 0)
		       | CP (Constpool.CPutf8 "F") => Vfloat (Real32.fromReal 0.0)
		       | CP (Constpool.CPutf8 "D") => Vdouble (Real64.fromReal 0.0)
		       | CP (Constpool.CPutf8 "Z") => Vint (Int32.fromInt 0)
		       | CP (Constpool.CPutf8 _)   => Vobj 0
		       | _                         => raise (match_error "init_field'")
	val _ = if isInstance
	        then Array.update (#attrs field,8,VALUE value)
		else if   List.exists isStatic (#flags field)
	             then Array.update (#attrs field,8,VALUE value)
	             else ()
    in  field
    end

fun init_field pool isInstance opField =
    case opField of
         NONE     => NONE
       | SOME fld => SOME (init_field' pool isInstance fld)



(********************************************)
(***** FUNCTIONS FOR BYTECODE EXECUTION *****)
(********************************************)
 
fun ireturn (ENV env_rec) =
    let val stack = #stack env_rec
        val current_frame = hd stack
	val invoking_frame = (hd o tl) stack
        val (newFrame,item) = popF current_frame
	val invoking_frame = pushF item invoking_frame
    in  ENV { method_area    = #method_area env_rec
            , method_counter = #method_counter env_rec
	    , heap           = #heap env_rec
	    , stack          = invoking_frame :: ((tl o tl) stack)
	    , pc_register    = #pc_register env_rec }
    end

fun store env n =
let val (newFrame,item) = popF (getCurrentFrame env)
    val locals = getLocals newFrame
    val _      = Array.update (locals,n,SOME item)
in  updateLocals newFrame locals
end

fun load env n =
let val frame  = getCurrentFrame env
    val locals = getLocals frame
    val item   = Array.sub (locals,n)
in  valOf item
end

fun aaload env =
    let val frame = getCurrentFrame env
        val (frame,idx) = case (popF frame) of
	                       (a,Sint b) => (a,b)
			     | _          => raise (match_error "aaload-1")
        val (frame,s) = case (popF frame) of
	                     (a,Sreference b) => (a,b)
			   | _                => raise (match_error "aaload-2")
	val heap = getHeap env
	val array = Heap.extract (heap,s)
	val item = Heap.arraySub (array,Int32.toInt idx)
    in  Sreference item
    end

fun arraylength env =
    let val frame = getCurrentFrame env
        val (frame,s) = case (popF frame) of
	                     (a,Sreference b) => (a,b)
			   | _                => raise (match_error "aaload-2")
	val heap = getHeap env
	val array = Heap.extract (heap,s)
	val item = Heap.arrayLength array
    in  Sint (Int32.fromInt item)
    end

fun aastore env = 
    let val frame = getCurrentFrame env
        val (frame,value) = case (popF frame) of
	                         (a,Sreference b) => (a,b)
			       | _                => raise (match_error "aastore-1")
        val (frame,idx)   = case (popF frame) of
	                         (a,Sint b) => (a,b)
			       | _          => raise (match_error "aastore-2")
        val (frame,s)     = case (popF frame) of
	                         (a,Sreference b) => (a,b)
			       | _                => raise (match_error "aastore-3")
	val heap = getHeap env
	val array = Heap.extract (heap,s)
	val _ = Heap.arrayUpdate (array,Int32.toInt idx, value)
    in  updateFrameAndPC env frame 1
    end

fun iadd env =
    let val frame = getCurrentFrame env
        val (frame1,i1) = case (popF frame) of
	                       (a,Sint b) => (a,b)
			     | _          => raise (match_error "iadd-1")
        val (frame2,i2) = case (popF frame1) of
	                       (a,Sint b)       => (a,b)
			     | _                => raise (match_error "iadd-2")	  
	val i3 = Int32.add i1 i2
    in  pushF (Sint i3) frame2
    end

fun imul env =
    let val frame = getCurrentFrame env
        val (frame1,i1) = case (popF frame) of
	                       (a,Sint b) => (a,b)
			     | _          => raise (match_error "imul-1")
        val (frame2,i2) = case (popF frame1) of
	                       (a,Sint b) => (a,b)
			     | _          => raise (match_error "imul-2")	  
	val i3 = Int32.multiply i1 i2
    in  pushF (Sint i3) frame2
    end

fun isub env =
    let val frame = getCurrentFrame env
        val (frame1,i1) = case (popF frame) of
	                       (a,Sint b) => (a,b)
			     | _          => raise (match_error "isub-1")
        val (frame2,i2) = case (popF frame1) of
	                       (a,Sint b) => (a,b)
			     | _          => raise (match_error "isub-2")	  
	val i3 = Int32.subtract i2 i1
    in  pushF (Sint i3) frame2
    end

fun putfield env idx =
    let val pool = getCurrentConstpool env
        val fld_name = case (lookup pool idx) of
	                    RCP (RCPfieldref {index,fieldIndex}) => extractNameFromNameTypeIdx pool index
	                  | CP (Constpool.CPfieldref {class,nameType}) => extractNameFromNameTypeIdx pool (cp2rcp nameType) 
			  | _  => raise (match_error "putfield-1")
        val frame = getCurrentFrame env
        val (frame,value) = popF frame        
	val (frame,hptr) = case (popF frame) of
	                           (f,Sreference p) => (f,p)
				 | _                => raise (match_error "putfield-3")
	
	val heap = getHeap env
	val new_heap = Heap.updateFieldValue (heap,stackItem2value value,hptr) fld_name
	val new_env = updateHeap env new_heap
    in  updateFrameAndPC new_env frame 3
    end

fun getfield env idx = 
    let val pool = getCurrentConstpool env
        val fld_name = case (lookup pool idx) of
	                    RCP (RCPfieldref {index,fieldIndex}) => extractNameFromNameTypeIdx pool index
	                  | CP (Constpool.CPfieldref {class,nameType}) => extractNameFromNameTypeIdx pool (cp2rcp nameType) 
		  	  | _                   => raise (match_error "getfield-1")
        val frame = getCurrentFrame env
        val (frame,obj_ref) = case (popF frame) of
	                           (f,Sreference p) => (f,p)
				 | _                => raise (match_error "getfield-2")
	val heap = getHeap env
	val (RuntimeClass.IDX class,hash) = case (Heap.extract (heap,obj_ref)) of
	                 Heap.Instance {class, hash} => (class,hash)
		       | _  => raise (match_error "getfield-3")
	val field = Polyhash.find hash fld_name
	val v = #value field
    in  updateFrameAndPC env (pushF (value2stackItem v) frame) 3
    end

fun if_icmp env cmp offset =
    let val frame = getCurrentFrame env
        val (frame,i1) = case (popF frame) of
	                      (a,Sint b) => (a,b)
			    | _          => raise (match_error "if_icmp-1")
	val (frame,i2) = case (popF frame) of
	                      (a,Sint b) => (a,b)
			    | _          => raise (match_error "if_icmp-2")
	val pc = if cmp (i2,i1)
	         then offset
		 else 3
    in  updateFrameAndPC env frame pc
    end

fun if_acmp env cmp offset =
    let val frame = getCurrentFrame env
        val (frame,p1) = case (popF frame) of
	                      (a,Sreference (Heap.HPTR b)) => (a,b)
			    | _          => raise (match_error "if_acmp-1")
	val (frame,p2) = case (popF frame) of
	                      (a,Sreference (Heap.HPTR b)) => (a,b)
			    | _          => raise (match_error "if_acmp-2")
	val pc = if p1 = p2
	         then offset
		 else 3
    in  updateFrameAndPC env frame pc
    end

fun if_default env cmp offset =
    let val frame = getCurrentFrame env
        val (frame,i) = case (popF frame) of
	                     (a,Sint b) => (a,b)
			   | _          => raise (match_error "if_default")
	val pc = if cmp (i,Int32.fromInt 0)
	         then offset
		 else 3
    in  updateFrameAndPC env frame pc
    end

fun anewarray env idx =
    let val pool = getCurrentConstpool env
        val class_name = extractClassNameFromPool pool idx
	val heap = getHeap env
	val frame = getCurrentFrame env
	val (frame,item) = popF frame
	val frame = pushF (idx2stackItem idx env) frame
	val s = case item of
	             Sint i => Int32.toInt i
		   | _      => raise (match_error "anewarray")
        val new_env = updateHeap env
		                  ( Heap.insert ( heap
				                , Heap.DataArray { object_type = class_name
					                             , array       = Array.array (s,Heap.HPTR 0) }
				                , ("[" ^ class_name) : string ) )
    in  updateFrameAndPC new_env frame 3
    end

fun iinc env (var,i) =
    let val frame    = getCurrentFrame env
        val locals   = getLocals frame
        val item     = case ((valOf o Array.sub) (locals,var)) of
	                    Sint a => a
			  | _ => raise (match_error "iinc")
	val new_item = Int32.add (Int32.fromInt i) item
	val _        = Array.update (locals,var,SOME (Sint new_item))
    in  updateLocals frame locals
    end



(******************************************)
(***** FUNCTIONS FOR NATIVE EXECUTION *****)
(******************************************)


fun println env n newline =
    let val frame = getCurrentFrame env
        val heap = getHeap env
        val arg   = case ((valOf o Array.sub) (getLocals frame,1)) of
	                 Environment.Sreference i => Heap.extract (heap,i)
		       | _      => raise (match_error "println-1")
	val s     = case arg of
	                 Heap.String s => s
		       | _        => raise (match_error "println-2")
	val _ = do_debug 10 s
	val _ = if   newline
	        then do_debug 10 "\n"
		else ()
    in  incrCurrentPC env 1
    end

fun int2string env i =
    let val frame = getCurrentFrame env
        val arg   = case ((valOf o Array.sub) (getLocals frame,0)) of
		         Environment.Sint i => (Int.toString o Int32.toInt) i
		       | _      => raise (match_error "int2string")
	val item  = (Sreference o Heap.HPTR o Heap.getRIndex o getHeap) env

	val frame = pushF item frame
	val new_env = updateFrameAndPC env frame i
        val new_env = updateHeap new_env
		                  ( Heap.insert ( getHeap new_env
					 	, Heap.String arg
						, "Ljava/lang/String;"))
    in  new_env
    end

fun parseInt env i =
    let val heap = getHeap env
        val frame = getCurrentFrame env
        val arg = case ((valOf o Array.sub) (getLocals frame,0)) of
		         Environment.Sreference i => Heap.extract (heap,i)
		       | _      => raise (match_error "parseInt")
	val (s,item) = case arg of
	                Heap.String s => (s,(valOf o Int.fromString) s)
		      | _      => raise (match_error "parseInt")
	val frame = pushF ((Sint o Int32.fromInt) item) frame
	val new_env = updateFrameAndPC env frame i
    in  new_env
    end



(*******************************)
(***** MAIN EXECUTION PART *****)
(*******************************)

local open Bytecode2
in 
   fun putstatic env idx =
       let val m_area = getMethodArea env
           val current_classname = extractCurrentClassName env
           val frame  = getCurrentFrame env
           val pool   = getPool m_area frame
           val (cl_idx, f_idx)  = case (lookup pool idx) of
	                               RCP (RCPfieldref {index,fieldIndex}) => (getFrameClassIdx frame, FIDX fieldIndex)
	                             | CP (Constpool.CPfieldref {class,nameType}) =>
			   	          let val fieldname = extractNameFromNameTypeIdx pool (cp2rcp nameType)
		          		      val classname = extractClassNameFromPool pool (cp2rcp class)
		          		      val new_env = try_or_load env (cp2rcp class) current_classname
					      val class_idx = getClassIdx new_env pool (cp2rcp class)
					          handle _ => raise (Error.noSuchClassError classname)
					      val result = string2findex m_area ( class_idx, fieldname)
						  handle _ => raise (Error.noSuchFieldError classname fieldname)
					  in  ( class_idx
				              , result )
				          end
			             | _                   => raise (match_error "putstatic")
	   val (frame,v) = popF frame
	   val value     = VALUE (stackItem2value v)
           val _         = updateFieldByIndex m_area (cl_idx,f_idx) value
       in frame 
       end

   and getstatic env idx =
       let val m_area = getMethodArea env
           val current_classname = extractCurrentClassName env
           val frame  = getCurrentFrame env
           val pool   = getPool m_area frame
           val (cl_idx, f_idx)  = case (lookup pool idx) of
	                               RCP (RCPfieldref {index,fieldIndex}) => (getFrameClassIdx frame, FIDX fieldIndex)
	                             | CP (Constpool.CPfieldref {class,nameType}) =>
				          let val fieldname = extractNameFromNameTypeIdx pool (cp2rcp nameType)
		          		      val classname = extractClassNameFromPool pool (cp2rcp class)
					      val new_env = try_or_load env (cp2rcp class) current_classname

					      val class_idx = getClassIdx new_env pool (cp2rcp class)
					          handle _ => raise (Error.noSuchClassError classname)
					     
					      val class_name = extractClassNameFromPool pool (cp2rcp class)
					      val result = string2findex m_area ( class_idx, fieldname )
						  handle _ => raise (Error.noSuchFieldError classname fieldname)
				          in  ( class_idx
				              , result )
				          end
			             | _                   => raise (match_error "getstatic")
           val stack_item  =  getFieldValueByIndex m_area (cl_idx,f_idx)
       in  value2stackItem stack_item
       end

    
   and getFieldArrays env (RuntimeClass.IDX i) =
       let val ma = getMethodArea env
           val method_array = #array ma
           val class = (valOf o Array.sub) (method_array, i)
	   val pool = #pool class
	   val fields_array = (#array o #fields) class
	   val size = ref 0
	   fun countInstanceField NONE     = ()
	     | countInstanceField (SOME _) = size := !size + 1
           val member_array = Array.array (Array.length fields_array,NONE)
	   val _ = Array.copy { src = fields_array
	                      , si = 0
	  		      , len = NONE
			      , dst = member_array
			      , di = 0 }
	   fun removeStaticField NONE = NONE
	     | removeStaticField (SOME f) = if   List.all isNonStatic (#flags f)
	                                    then SOME f
			                    else NONE
	   val _ = Array.modify removeStaticField member_array
	   val _ = Array.app countInstanceField member_array
	   fun getVal (VALUE v) = v
	     | getVal _         = raise (match_error "getFieldArrays")
	   fun member2field { flags
	                    , name
		      	    , desc
			    , attrs } = { flags = flags
			                , name  = extract_name pool name
				        , desc  = extract_name pool desc
				        , value = (getVal o Array.sub) (attrs,RuntimeClass.value) } : Heap.field
	   val new_array = Array.array (!size,NONE)
	   val counter = ref 0
	   val new_counter = ref 0
	   val _ = while (!counter < Array.length member_array)
	              do let val member = Array.sub (member_array,!counter)
	  	         in
		         ( (   if   member <> NONE
		               then ( Array.update ( new_array
			                           , !new_counter
			  	  		   , (SOME o member2field) (init_field' pool true (valOf member)));
		                      new_counter := !new_counter + 1
			            )
			       else ()
		           ) ; counter := !counter + 1
		         )
		         end
	
           fun g classname env idx =
	       let val name = classname ^ ".class"
	           val new_env = classload name env (RuntimeClass.IDX i,idx)
		   val new_ma = getMethodArea new_env
		   val idx2 = label2index new_ma (LBL classname)
	       in  (idx2,new_env)
	       end
	
	   val (rest,new_env) = case (#super class) of
	                             SOME n => let val class_name = extractClassNameFromPool pool n
	 		                           val (idx,new_env) = ( case (label2index_try ma (LBL class_name)) of
			                                                      RuntimeClass.IDX 0 => g class_name env n
				                                            | x                  => (x,env)
				                                       )
				               in  getFieldArrays new_env idx
				               end
		                   | NONE   => ( [], env )
       in  ( rest @ [new_array], new_env )
       end

   and createInstanceHash env i =
       let val instanceHash = Polyhash.mkPolyTable (100, find_error "instance_hash")
           fun f x = let val y = valOf x
	             in  Polyhash.insert instanceHash (#name y ^ #desc y, y)
		     end
           fun createInstanceHash' x = Array.app f x
           val (arrays,new_env) = getFieldArrays env (RuntimeClass.IDX i)
           val _ = List.app createInstanceHash' arrays
       in  (instanceHash,new_env)
       end
 
   and getInstanceFields env i =
       let val (new_hash,new_env) = createInstanceHash env i
       in  ( { class = RuntimeClass.IDX i
  	     , hash  = new_hash }
	   , new_env )
       end

   and addHeapItem_option env idx =
       let val heap = getHeap env           
           val current_classname = extractCurrentClassName env
           val pool = getCurrentConstpool env
       in  case (lookup pool idx) of
                CP (Constpool.CPstring idx2) =>
	           ( case (lookup pool (cp2rcp idx2)) of
	                  CP (Constpool.CPutf8 s) =>
			         updateHeap env
		                            ( Heap.insert ( heap
				 	 	          , Heap.String s
						          , "Ljava/lang/String;") )
                        | _                       => raise (match_error "addHeapItem_option-1")
		   ) 
              | CP (Constpool.CPdouble d) => env
              | CP (Constpool.CPfloat f) => env 
              | CP (Constpool.CPint i) => env 
              | CP (Constpool.CPlong l) => env
              | RCP (RCPclassref {index,classIdx}) =>
                    let val (instanceFields,new_env) = getInstanceFields env classIdx
	                val classname                = extract_name pool index
	                val objectPool               = getConstpool new_env classname
	            in  updateHeap new_env
		                  ( Heap.insert ( heap
				                , Heap.Instance instanceFields
				                , classname ) )
	            end
              | CP (Constpool.CPclass index) =>
                    let val new_env                  = try_or_load env idx current_classname
		        val clname = extractClassNameFromPool pool idx
			val ma = getMethodArea env
			val (RuntimeClass.IDX xx) = label2index ma (LBL clname)
		        val (instanceFields,new_env) = getInstanceFields env xx
	                val classname                = extract_name pool (cp2rcp index)
	                val objectPool               = getConstpool new_env classname
	            in  updateHeap new_env
		                  ( Heap.insert ( heap
				                , Heap.Instance instanceFields
				                , classname ) )
	            end
              | _  => raise (match_error "addHeapItem_option-2")
       end
   
   (*merge these two functions into one*)
   and try_or_load env idx current_classname =
      let val pool = getConstpool env current_classname
          val classname = extractClassNameFromPool pool idx
          val name = classname ^ ".class"
          val method_hash = (#hash o getMethodArea) env
          val class_index = label2index (getMethodArea env) (LBL current_classname)
      in  if   try (method_hash,classname)
          then classload name env (class_index,idx) (*update current constpool & method_area*)
          else env
      end
      
   and new env idx =
       let val current_classname = extractCurrentClassName env
           val new_env = try_or_load env idx current_classname
	   val frame = getCurrentFrame new_env
	   
	   val frame = pushF (idx2stackItem idx new_env) frame
	   val new_env = updateFrameAndPC new_env frame 3
	   val new_env  = addHeapItem_option new_env idx
       in  new_env
       end

   and ex_try2 env class_name method_name (idx,current_classname) =
       let val new_env = try_or_load env idx current_classname
           val method_area = getMethodArea new_env
           val class_index = Polyhash.find (#hash method_area) class_name
	   val class = (valOf o Array.sub) (#array method_area,class_index)
	   val methods = #methods class
	   val result = SOME (Polyhash.find (#hash methods) method_name) handle _ => NONE
	   fun toBool NONE     = false
	     | toBool (SOME _) = true
       in  (toBool result,new_env)
       end

   and findAndExecuteMethod e cn (currentClassName,requestedClassIdx) (classname,methodname,isInstance,args) =
       let val (bo, new_e) = ex_try2 e cn methodname (requestedClassIdx,currentClassName)
	   val new_c = findClassByLabel (getMethodArea new_e) (LBL cn)
       in  if   bo
           then (execute' new_e
	                  cn
			  (methodname,args)
			  isInstance )
	   else if   option2bool (#super new_c) (*has_super?*)
	        then let val requestedClassIdx = (valOf o #super) new_c
		         val new_cn  = extractClassNameFromPool (#pool new_c) requestedClassIdx
                     in  findAndExecuteMethod new_e
					      new_cn
					      (cn,requestedClassIdx)
					      (classname,methodname,isInstance,args)
		     end
		else raise (Error.noSuchMethodError classname methodname)
       end

   and invoke env idx isInstance =
       let val pool =  getCurrentConstpool env
	   val currentClassName = extractCurrentClassName env
	   val currentClassIdx = extractCurrentClassIdx env

           val aa = case (lookup pool (cp2rcp idx)) of
	                 RCP (RCPmethodref i) =>
		             let val methods = getCurrentMethods env
			         val method  = valOf (Array.sub (#array methods, i))
			         val name_string = extract_name pool (#name method)
			         val type_string = extract_name pool (#desc method)
				 val name = name_string ^ type_string
			     in  execute' env currentClassName (name,[]) isInstance
			     end
		       | CP (Constpool.CPmethodref {class,nameType}) =>
		            let val new_env = try_or_load env (cp2rcp class) currentClassName
			        fun f name desc = extract_name pool (cp2rcp name) ^ extract_name pool (cp2rcp desc)
			    in  ( case (lookup pool (cp2rcp nameType)) of
		                         CP (Constpool.CPnametype {name,desc}) =>
				            let val classname = extractClassNameFromPool pool (cp2rcp class)
					        val methodname = f name desc
					    in findAndExecuteMethod new_env 
					                            classname
								    (currentClassName,currentClassIdx)
								    (classname,methodname,isInstance,[])
					    end
			               | _ => raise (match_error "invoke-1") )
			    end
		       | _ => raise (match_error "invoke-2")
       in  aa
       end

   and ex_instr env =
       let val frame = getCurrentFrame env
           val asize = (Int.toString o Array.length) (getCode frame)
	   val addr = addrValue (getPC frame)
           val instr = Array.sub (getCode frame, addr ) 
       in  case instr of
                Jbipush byte   => (ex_instr ( updateFrameAndPC   env
		                                                (pushF (Sint (Int32.fromInt byte))
								       frame )
								2 ))
              | Jsipush short  => (ex_instr ( updateFrameAndPC  env
		                                                (pushF (Sint (Int32.fromInt short))
								       frame )
								3 ))
              | Jaconst_null    => (ex_instr (  updateFrameAndPC  env
		                                                 (pushF (Sreference (Heap.HPTR 0))
				 				        frame )
								 1 ))
              | Jiconst_0      => (ex_instr (  updateFrameAndPC  env
		                                                (pushF (Sint (Int32.fromInt 0))
								       frame )
								1 ))
              | Jiconst_1      => (ex_instr (  updateFrameAndPC  env
		                                                (pushF (Sint (Int32.fromInt 1))
								       frame )
								1 ))
              | Jiconst_2      => (ex_instr (  updateFrameAndPC  env
		                                                (pushF (Sint (Int32.fromInt 2))
								       frame )
								1 ))
              | Jiconst_3      => (ex_instr (  updateFrameAndPC  env
		                                                (pushF (Sint (Int32.fromInt 3))
								       frame )
								1 ))
              | Jiconst_4      => (ex_instr (  updateFrameAndPC  env
		                                                (pushF (Sint (Int32.fromInt 4))
								       frame )
								1 ))
              | Jiconst_5      => (ex_instr (  updateFrameAndPC  env
		                                                (pushF (Sint (Int32.fromInt 5))
								       frame )
								1 ))
              | Jastore local_idx => (ex_instr (  updateFrameAndPC  env
		                                                (store env (Localvar.toInt local_idx))
								2 ))
              | Jastore_0      => (ex_instr (  updateFrameAndPC  env
		                                                (store env 0)
								1 ))
              | Jastore_1      => (ex_instr (  updateFrameAndPC  env
		                                                (store env 1)
								1 ))
              | Jastore_2      => (ex_instr (  updateFrameAndPC  env
		                                                (store env 2)
								1 ))
              | Jastore_3      => (ex_instr (  updateFrameAndPC  env
		                                                (store env 3)
								1 ))
              | Jistore local_idx => (ex_instr ( updateFrameAndPC env
		                                                  (store env (Localvar.toInt local_idx))
								  2 ))
              | Jistore_0      => (ex_instr (  updateFrameAndPC env
		                                                (store env 0)
								1 ))
              | Jistore_1      => (ex_instr (  updateFrameAndPC  env
		                                                (store env 1)
								1 ))
              | Jistore_2      => (ex_instr (  updateFrameAndPC  env
		                                                (store env 2)
								1 ))
              | Jistore_3      => (ex_instr (  updateFrameAndPC  env
		                                                (store env 3)
								1 ))
              | Jaastore      => (ex_instr (  aastore env ))
              | Jarraylength  => (ex_instr (  updateFrameAndPC env
		                                              (pushF (arraylength env)
								     frame )
							      1 ))
              | Jaaload       => (ex_instr (  updateFrameAndPC env
		                                              (pushF (aaload env)
								     frame )
							      1 ))
              | Jaload local_idx => (ex_instr (  updateFrameAndPC env
		                                                  (pushF (load env (Localvar.toInt local_idx))
								         frame )
							          2 ))
              | Jaload_0      => (ex_instr (  updateFrameAndPC env
		                                               (pushF (load env 0)
								      frame )
							       1 ))
              | Jaload_1      => (ex_instr (  updateFrameAndPC env
		                                               (pushF (load env 1)
								      frame )
							       1 ))
              | Jaload_2      => (ex_instr (  updateFrameAndPC env
		                                               (pushF (load env 2)
								      frame )
							       1 ))
              | Jaload_3      => (ex_instr (  updateFrameAndPC env
		                                               (pushF (load env 3)
								      frame )
							       1 ))
              | Jiload local_idx => (ex_instr (  updateFrameAndPC  env
		                                                   (pushF (load env (Localvar.toInt local_idx))
								          frame )
							           2 ))
              | Jiload_0      => (ex_instr (  updateFrameAndPC  env
		                                               (pushF (load env 0)
								      frame )
							       1 ))
              | Jiload_1      => (ex_instr (  updateFrameAndPC  env
		                                               (pushF (load env 1)
								      frame )
							       1 ))
              | Jiload_2      => (ex_instr (  updateFrameAndPC  env
		                                               (pushF (load env 2)
								      frame )
							       1 ))
              | Jiload_3      => (ex_instr (  updateFrameAndPC  env
		                                               (pushF (load env 3)
								      frame )
							       1 ))
	      | Jputfield idx => (ex_instr (putfield env (cp2rcp idx) ))
	      | Jgetfield idx => (ex_instr (getfield env (cp2rcp idx) ))
              | Jputstatic idx => (ex_instr (  updateFrameAndPC env
	                                                        (putstatic env (cp2rcp idx) )
								3 ))
              | Jgetstatic idx => (ex_instr ( updateFrameAndPC  env
		                                                (pushF (getstatic env (cp2rcp idx) )
								       frame )
							        3 ))
              | Jldc idx => let val frame = pushF (idx2stackItem (cp2rcp idx) env) frame
				val env2  = addHeapItem_option env (cp2rcp idx)
			    in  (ex_instr (  updateFrameAndPC env2 frame 2 ))
			    end
              | Jiadd          => (ex_instr ( updateFrameAndPC env (iadd env) 1 ))
              | Jimul          => (ex_instr ( updateFrameAndPC env (imul env) 1 ))
              | Jisub          => (ex_instr ( updateFrameAndPC env (isub env) 1 ))
	      | Jpop => (ex_instr (updateFrameAndPC env ((#1 o popF o getCurrentFrame) env) 1 ))
	      | Jdup => (ex_instr (updateFrameAndPC env ((dup o getCurrentFrame) env) 1 ) )
	      | Jnew idx => (ex_instr (new env (cp2rcp idx) ))
	      | Janewarray idx => (ex_instr (anewarray env (cp2rcp idx) ))
              | Jinvokevirtual idx => (ex_instr ( invoke (incrCurrentPC env 3) idx true ) )
              | Jinvokespecial idx => (ex_instr ( invoke (incrCurrentPC env 3) idx true ) )
              | Jinvokeinterface idx => (ex_instr ( invoke (incrCurrentPC env 3) idx true ) )
              | Jinvokestatic idx => (ex_instr ( invoke (incrCurrentPC env 3) idx false ))
	      | Jreturn        => (popCurrentFrame env)
	      | Jireturn       => ( ireturn env )
	      | Jareturn       => ( ireturn env ) 
	      | Jif_icmpne (OFF i) => (ex_instr ( if_icmp env icmp_ne i ))
	      | Jif_icmple (OFF i) => (ex_instr ( if_icmp env icmp_le i ))
	      | Jif_icmplt (OFF i) => (ex_instr ( if_icmp env icmp_lt i ))
	      | Jif_icmpgt (OFF i) => (ex_instr ( if_icmp env icmp_gt i ))
	      | Jif_icmpge (OFF i) => (ex_instr ( if_icmp env icmp_ge i ))
	      | Jif_icmpeq (OFF i) => (ex_instr ( if_icmp env icmp_eq i ))
	      | Jif_acmpne (OFF i) => (ex_instr ( if_acmp env op<> i ))
	      | Jif_acmpeq (OFF i) => (ex_instr ( if_acmp env op= i ))
	      | Jgoto (OFF i) => (ex_instr ( incrCurrentPC env i ))
	      | Jgoto_w (OFF i) => (ex_instr ( incrCurrentPC env i ))
	      | Jiinc {var,const} => (ex_instr ( updateFrameAndPC  env
		                                                (iinc env ((Localvar.toInt var),const))
								3 ))
	      | Jifne (OFF i) => (ex_instr ( if_default env op<> i ))
	      | Jifeq (OFF i) => (ex_instr ( if_default env op= i ))
	      | Nprintln => (ex_instr ( println env 1 true ))
	      | Nprint => (ex_instr ( println env 1 false ))
	      | Nint2string => (ex_instr ( int2string env 1 ))
	      | NparseInt => (ex_instr ( parseInt env 1 ))
              | _              => raise (unimplemented_error "ex_instr")
       end

and execute (ENV env_rec) class (method,topArgs) =
    execute' (ENV env_rec) class (method,topArgs) false

and execute' (ENV env_rec) class (method,topArgs) isInstance =
    let val top_stack          = #stack env_rec
        val m_area             = #method_area env_rec
	val cl_idx             = label2index m_area (LBL class)
        val current_class      = findClassByIndex m_area cl_idx
        val current_cp         = #pool current_class

        val heap               = ref (#heap env_rec)

	val cm_idx             = string2mindex m_area (cl_idx,method)
	                             handle _ => raise (Error.noSuchMethodError class method)
	val current_method     = findMethodByIndex m_area (cl_idx,cm_idx)

	val current_code       = Array.sub ((#attrs current_method),code)
	val code_record        = case current_code of
	                              RuntimeClass.CODE r => r
				    | _      => raise (match_error "execute'-1")
	val instrs             = #code code_record
	
	val limit_locals       = #locals code_record
        val local_vars         = Array.array (limit_locals, NONE)
        val (args, result)     = Parser.parseTypes (extract_name current_cp (#desc current_method) ) 
	val nr_of_args         = if isInstance
	                         then List.length args + 1
				 else List.length args

        val top_stack = if nr_of_args = 0
                        then top_stack
		        else if method = "main([Ljava/lang/String;)V" (*topArgs <> []*)
			     then let val (ptrList,newHeap) = List.foldr arg2item ([],!heap) topArgs
			              val _  = heap := newHeap
				      val sa = Heap.DataArray { object_type = "Ljava/lang/String;"
				                              , array       = Array.fromList ptrList }
				      val _  = Array.update( local_vars
			                                   , 0
						 	   , SOME ( ( Environment.Sreference o
							              Heap.HPTR o
								      Heap.getRIndex
								    ) (!heap)
								  )
							   )
				      val newHeap  = Heap.insert (!heap,sa,"[Ljava/lang/String;")
				      val _        = heap := newHeap
			          in top_stack
				  end
			     else let val parent_frame = pop top_stack
                                      val count = ref (nr_of_args-1)
	                              val pf = ref parent_frame
	                              val _ = while (!count >= 0) do
	                                      let val (new_frame,var) =  popF (!pf)
					          val ds = case var of
						                Sint i => "int" ^ (Int.toString o Int32.toInt) i
							      | Sreference (Heap.HPTR h) => "ref" ^ Int.toString h
							      | _ => "unknown"
					      in  ( (Array.update (local_vars, !count, SOME var));
		                                    pf    := new_frame;
		                                    count := !count - 1 )
		                              end
                                      val new_parent_frame   = !pf
	                          in  push (new_parent_frame,tl top_stack)
			          end
				
        val current_frame      = FRAME { stack          = []
	                               , local_vars     = local_vars
				       , current_class  = cl_idx
				       , current_method = cm_idx
				       , current_code   = instrs
				       , current_pc     = NATIVE 0 }

	val top_stack      = push (current_frame,top_stack)
        val (ENV new_env) = ex_instr ( ENV { method_area    = m_area
                                           , method_counter = #method_counter env_rec 
          	                           , heap           = !heap
	                                   , stack          = top_stack
	                                   , pc_register    = #pc_register env_rec }  )

    in  (ENV new_env)
    end

and classload cf env (current_class_index,idx) =
    let val class   = (Transform.file2run o Parser.parse) cf handle _ => raise (Error.NoClassDefFoundError cf) 
	val cl_name = RuntimePool.extract_class (#pool class) (#this class)
	val ma      = getMethodArea env
	val mc      = getMethodCounter env
	val _       = Array.update ( #array ma
	                           , mc
				   , SOME class )
	val _       = Polyhash.insert (#hash ma) (cl_name, mc)
	val pool    = findCPByIndex ma current_class_index
	val pool2   = #pool class


	val _          = Array.modify (init_field pool2 false) ((#array o #fields) class)

	val i = case (lookup pool idx) of
	             CP (Constpool.CPclass a) => a
		   | _                        => raise (match_error "classload") 
        val env2 = updateConstpool ( ENV { method_area    = ma
                      				, method_counter = mc + 1
	              				, heap           = getHeap env
	              				, stack          = getEnvStack env
	              				, pc_register    = getEnvPC env } )
		      			  ( RuntimePool.indexValue idx , RCP (RCPclassref { index = cp2rcp i
											  , classIdx = mc}) )
				   current_class_index
        val has_clinit = ex_try env2 cl_name "<clinit>()V"
	val env3 = if   has_clinit = true
	           then execute env2 cl_name ("<clinit>()V",[])
		   else env2
    
    in  env3
    end

fun bootload' cf args debug =
    let val _ = no_debugging := debug
    
        val m_hash     = Polyhash.mkPolyTable (50,find_error "method_hash")
        val m_array    = Array.array (50,NONE)
        val class      = (Transform.file2run o Parser.parse) cf handle _ => raise (Error.NoClassDefFoundError cf)
        val pool       = #pool class
	val this_idx   = #this class
	val _          = Array.modify (init_field pool false) ((#array o #fields) class)

	val i = case (lookup pool this_idx) of
	             CP (Constpool.CPclass a) => a
		   | _                        => raise (match_error "bootload")

        val cl_name    = RuntimePool.extract_class pool this_idx
	
	val _          = Array.update (m_array,1, SOME object_class)
	val _          = Polyhash.insert m_hash ("java/lang/Object", 1)
	
	val _          = Array.update (m_array,2, SOME print_class)
	val _          = Polyhash.insert m_hash ("java/io/PrintStream", 2)
	
	val _          = Array.update (m_array,3, SOME printstream_class)
	val _          = Polyhash.insert m_hash ("java/lang/System", 3)

        val _          = Array.update (m_array,4, SOME integer_class)
	val _          = Polyhash.insert m_hash ("java/lang/Integer", 4)

	val _          = Array.update (m_array,5, SOME class)
	val _          = Polyhash.insert m_hash (cl_name, 5)	
	
	val env1       = ENV { method_area    = {array=m_array,hash=m_hash}
	                     , method_counter = 6
	                     , heap           = Heap.newHeap
			     , stack          = []
			     , pc_register    = NATIVE 0 }
	val env2       = updateConstpool env1
	                                 ( RuntimePool.indexValue this_idx
	                                 , RCP ( RCPclassref { index = cp2rcp i
							     , classIdx = 5 }
					 ))
					 (RuntimeClass.IDX 5)
	val has_clinit = ex_try env2 cl_name "<clinit>()V"
	val env3 = if   has_clinit = true
	           then execute env2 cl_name ("<clinit>()V",[])
		   else env2
    in  findAndExecuteMethod env3
                             cl_name
			     (cl_name,this_idx)
                             (cl_name,"main([Ljava/lang/String;)V",false,args)
    
    end



fun bootload cf args = bootload' cf args true

(*timed execute*)
fun tex file args =
   let val tr = Timer.startRealTimer ()
       val env = bootload file args
   in  Time.toMilliseconds (Timer.checkRealTimer tr)
   end


end (*of local*)

end (*of struct*)
