(*                                                                      *)
(* MLJVM - a Java Virtual Machine implemented in Standard ML            *)
(*                                                                      *)
(* http://homepages.inf.ed.ac.uk/stg/research/projects/MLJVM/           *)
(*                                                                      *)
(* Copyright (C) 2003 Stephen Gilmore and Laura Korte, The              *)
(* University of Edinburgh.                                             *)
(*                                                                      *)
(* This program is free software; you can redistribute it and/or        *)
(* modify it under the terms of the GNU General Public License          *)
(* as published by the Free Software Foundation; either version 2       *)
(* of the License, or (at your option) any later version.               *)
(*                                                                      *)
(* This program is distributed in the hope that it will be useful,      *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of       *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *)
(* GNU General Public License for more details.                         *)
(*                                                                      *)
(* You should have received a copy of the GNU General Public License    *)
(* along with this program; if not, write to the Free Software          *)
(* Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *)
(* MA 02111-1307, USA.                                                  *)
(*                                                                      *)

structure Heap =
struct

(**********************)
(***** DEBUG INFO *****)
(**********************)


val debug_string = ref "" : string ref
fun do_debug n ss = case n of
		              0 => ()
			    | 1 => ()
			    | 2 => ()
			    | _ => ((print ss);(TextIO.flushOut TextIO.stdOut))


(*********************************)
(***** EXCEPTION DEFINITIONS *****)
(*********************************)

local
   val file = "Heap.sml"
in
   fun error s = Error.error file s
   fun match_error s = Error.error file s
end



(*hpointer : data/object-area -> intermediate*)
datatype hpointer = HPTR of int
(*opointer : intermediate -> object*)
datatype opointer    = OPTR of int


type field = { flags : Modifiers.access_flag list
             , name  : string
	     , desc  : string
	     , value : RuntimeClass.value }

datatype object = String of string
		| DataArray of { object_type : string (*classname*)
		               , array       : hpointer Array.array }
		| Instance of { class : RuntimeClass.index
			      , hash  : (string,field) Polyhash.hash_table }
                | DUMMY (*placeholder*)
		     
type object_area = object Array.array
fun extractObj objs (OPTR i) = Array.sub (objs,i)

fun arraySub (DataArray {object_type,array},n) = Array.sub (array,n)
  | arraySub _                                 = raise (match_error "arraySub")

fun arrayLength (DataArray {object_type,array}) = Array.length array
  | arrayLength _                               = raise (match_error "arrayLength")

fun arrayUpdate ( DataArray { object_type
                                  , array }
	           , n
		   , value ) = Array.update (array,n,value)
  | arrayUpdate _                             = raise (match_error "arrayUpdate")

(*intermediate heap area*)
datatype pair     = PAIR of (opointer * string) (*string = reference-by-name to object-class*)
                  | EMPTY (*placeholder*)
type refheap   = pair Array.array


(*the heap itself*)
datatype heap     = HEAP of { refheap     : refheap
                            , object_area : object_area
			    , r_index     : int ref 
			    , o_index     : int ref } 
fun getRefHeap    (HEAP {refheap,object_area,r_index,o_index}) = refheap
fun getObjectArea (HEAP {refheap,object_area,r_index,o_index}) = object_area
fun getRIndex     (HEAP {refheap,object_area,r_index,o_index}) = !r_index
fun getOIndex     (HEAP {refheap,object_area,r_index,o_index}) = !o_index

val newHeap = HEAP { refheap = Array.array(1000,EMPTY)
                   , object_area = Array.array(1000,DUMMY)
		   , r_index = ref 1
		   , o_index = ref 1 }

fun insert (HEAP heap_rec,object,class) = 
let val _ = Array.update(  #object_area heap_rec
		         , !(#o_index heap_rec)
		         , object )
    val _ = Array.update(  #refheap heap_rec
		         , !(#r_index heap_rec)
		         , PAIR ((OPTR o ! o #o_index) heap_rec, class) )
in  HEAP { refheap = #refheap heap_rec
         , object_area = #object_area heap_rec
	 , r_index = ref (! (#r_index heap_rec) + 1)
	 , o_index = ref (! (#o_index heap_rec) + 1) }
end

fun updateFieldValue (HEAP heap_rec, value, HPTR hptr) fieldName =
    let val o1 = Array.sub (#refheap heap_rec,hptr)
        val (optr,class) = case o1 of
                                PAIR (OPTR optr,class) => (optr,class)
			      | _                      => raise (match_error "updateFieldValue-1")
	val o2 = Array.sub (#object_area heap_rec, optr)
        val (class,hash) = case o2 of
	                        Instance {class, hash} => (class,hash)
	                      | String s => raise (match_error s)
			      | _                      => raise (match_error "updateFieldValue-2")
	val field = Polyhash.find hash fieldName
	val new_field = { flags = #flags field
	                , name  = #name field
			, desc  = #desc field
			, value = value }
	val _ = Polyhash.insert hash (fieldName,new_field)
	val newObject = Instance { class = class
				 , hash = hash }
        val _ = Array.update ( #object_area heap_rec
		             , optr
		             , newObject )
    in  HEAP { refheap = #refheap heap_rec
             , object_area = #object_area heap_rec
	     , r_index = #r_index heap_rec
	     , o_index = #o_index heap_rec }
    end

fun extract (heap,HPTR hptr) =
    let val o1 = Array.sub (getRefHeap heap,hptr)
        val (optr,class) = case o1 of
	                        PAIR (OPTR optr,class) => (optr,class)
			      | _                      => raise (match_error "extract")
    in  Array.sub (getObjectArea heap,optr)
    end


end;
