(*                                                                      *)
(* MLJVM - a Java Virtual Machine implemented in Standard ML            *)
(*                                                                      *)
(* http://homepages.inf.ed.ac.uk/stg/research/projects/MLJVM/           *)
(*                                                                      *)
(* Copyright (C) 2003 Stephen Gilmore and Laura Korte, The              *)
(* University of Edinburgh.                                             *)
(*                                                                      *)
(* This program is free software; you can redistribute it and/or        *)
(* modify it under the terms of the GNU General Public License          *)
(* as published by the Free Software Foundation; either version 2       *)
(* of the License, or (at your option) any later version.               *)
(*                                                                      *)
(* This program is distributed in the hope that it will be useful,      *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of       *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *)
(* GNU General Public License for more details.                         *)
(*                                                                      *)
(* You should have received a copy of the GNU General Public License    *)
(* along with this program; if not, write to the Free Software          *)
(* Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *)
(* MA 02111-1307, USA.                                                  *)
(*                                                                      *)

structure JasminPrint :> JasminPrint =
struct

(* NOTE - Not Supported: .field final <name> <type> = <value> *)

exception Error of string;
exception NoSourceFileSpecError;
exception FormatError;
exception UnsupportedModifierError of string;

open MyClassFile

(**********************)
(***** DEBUG INFO *****)
(**********************)

val debug_string = ref "" : string ref
fun do_debug 0 ss = ()
  | do_debug n ss =  ((print ss);(TextIO.flushOut TextIO.stdOut))



(* pretty_print flags of: classes, fields, methods *)
fun pp_modifiers access_flags kind =
let val mod_function = case kind of
                         "class"  => Modifiers.class_mod
                       | "method" => Modifiers.method_mod
		       | "field"  => Modifiers.method_mod
		       | s        => raise UnsupportedModifierError s
    val modifier_list = Modifiers.priority_list
    fun try_mod n (x::xs) = if n-x < 0
                            then ("",n,xs)
			    else (mod_function x ^ " ", n-x, xs)
      | try_mod n []      = raise Error "Cannot do pp_modifiers"
    fun compute_flags (mods_s,flags,l) = if flags > 0
                                         then ( compute_flags
					        o
						( fn ( new_mod         , new_flags, mod_list) =>
					             ( new_mod ^ mods_s, new_flags, mod_list)
						)
					      ) (try_mod flags l)
		                         else (mods_s,0)
    val (result,n)  = compute_flags ( "", access_flags, modifier_list )
in  result
end

(* pretty_print the .source statement *)
fun pp_srcfile l pool =
let fun isSourceFile (SRCFILE _) = true
      | isSourceFile  _          = false 
    fun strip_cons ( SOME (SRCFILE { attr = idx
                                   , file = idx2 }
		 	  )
		    ) = idx2
      | strip_cons  _ = raise NoSourceFileSpecError
    val getFileName   = (Constpool.extract_name pool) o strip_cons
    val srcfile       = getFileName (List.find isSourceFile l)
in ".source " ^ srcfile
end

(* pretty_print a .field statement *)
fun pp_field pool (mem:member) = 
let val flags = Word16.toInt (#flags mem)
    val mods  = pp_modifiers flags "field"
    val name  = Constpool.extract_name pool (#name mem) ^ " "
    val desc  = Constpool.extract_name pool (#desc mem) ^ " "
in  "\n.field " ^ mods ^ name ^ desc
end

(* pretty_print the body of a .method statement*)
local open Bytecode Jvmtype
   fun ppType Tboolean   = "Z"
     | ppType Tchar      = "C"
     | ppType Tfloat     = "F"
     | ppType Tdouble    = "D"
     | ppType Tbyte      = "B"
     | ppType Tshort     = "S"
     | ppType Tint       = "I"
     | ppType Tlong      = "J"
     | ppType (Tarray t) = "[" ^ ppType t
     | ppType (Tclass c) = "L" ^ className c

   fun ppType_option (SOME t) = ppType t
     | ppType_option  NONE    = "V"
   fun indexValue idx = String.extract(Localvar.toString idx, 4, NONE)
   fun labelValue idx = String.extract(Label.toString idx, 6, NONE)

   fun pp_instr instr =
       case instr of
        Jnop         => "\n\tnop"
      | Jaconst_null => "\n\taconst_null"
      | Jiconst_m1   => "\n\ticonst_m1"
      | Jiconst_0    => "\n\ticonst_0"
      | Jiconst_1    => "\n\ticonst_1"
      | Jiconst_2    => "\n\ticonst_2"
      | Jiconst_3    => "\n\ticonst_3"
      | Jiconst_4    => "\n\ticonst_4"
      | Jiconst_5    => "\n\ticonst_5"
      | Jlconst_0    => "\n\tlconst_0"
      | Jlconst_1    => "\n\tlconst_1"  (*10*)
      | Jfconst_0    => "\n\tfconst_0"
      | Jfconst_1    => "\n\tfconst_1"
      | Jfconst_2    => "\n\tfconst_2"
      | Jdconst_0    => "\n\tdconst_0"
      | Jdconst_1    => "\n\tdconst_1"
      | Jbipush    i => "\n\tbipush " ^ Int.toString i
      | Jsipush    i => "\n\tsipush " ^ Int.toString i
      | Jldc   c  => "\n\tldc " ^
                     ( case c of Cint i    => "<unimplemented int>"
		               | Cfloat f  => "<unimplemented int>"
		 	       | Cstring s => "\"" ^ s ^ "\""
			       | _         => raise Error "Jldc formaterror"
		     )
      | Jldc_w c  => "\n\tldc_w " ^
                     ( case c of Cint i    => "<unimplemented int>"
		               | Cfloat f  => "<unimplemented int>"
			       | Cstring s => "\"" ^ s ^ "\""
			       | _         => raise Error "Jldc_w formaterror"
		     )
      | Jldc2_w c => "\n\tldc2_w " ^
                     ( case c of Clong l    => "<unimplemented int>"
		               | Cdouble d  => "<unimplemented int>"
			       | _         => raise Error "Jldc2_w formaterror"
		     )                  (*20*)
      | Jiload  index => "\n\tiload " ^ indexValue index
      | Jlload  index => "\n\tlload " ^ indexValue index
      | Jfload  index => "\n\tfload " ^ indexValue index
      | Jdload  index => "\n\tdload " ^ indexValue index
      | Jaload  index => "\n\taload " ^ indexValue index
      | Jiload_0      => "\n\tiload_0"
      | Jiload_1      => "\n\tiload_1"
      | Jiload_2      => "\n\tiload_2"
      | Jiload_3      => "\n\tiload_3"
      | Jlload_0      => "\n\tlload_0"  (*30*)
      | Jlload_1      => "\n\tlload_1"
      | Jlload_2      => "\n\tlload_2"
      | Jlload_3      => "\n\tlload_3"
      | Jfload_0      => "\n\tfload_0"
      | Jfload_1      => "\n\tfload_1"
      | Jfload_2      => "\n\tfload_2"
      | Jfload_3      => "\n\tfload_3"
      | Jdload_0      => "\n\tdload_0"
      | Jdload_1      => "\n\tdload_1" 
      | Jdload_2      => "\n\tdload_2"  (*40*)
      | Jdload_3      => "\n\tdload_3"
      | Jaload_0      => "\n\taload_0"
      | Jaload_1      => "\n\taload_1"
      | Jaload_2      => "\n\taload_2"
      | Jaload_3      => "\n\taload_3"
      | Jiaload       => "\n\tiaload"
      | Jlaload       => "\n\tlaload"
      | Jfaload       => "\n\tfaload"
      | Jdaload       => "\n\tdaload"
      | Jaaload       => "\n\taaload"   (*50*)
      | Jbaload       => "\n\tbaload"
      | Jcaload       => "\n\tcaload"
      | Jsaload       => "\n\tsaload"
      | Jistore index => "\n\tistore " ^ indexValue index
      | Jlstore index => "\n\tlstore " ^ indexValue index
      | Jfstore index => "\n\tfstore " ^ indexValue index
      | Jdstore index => "\n\tdstore " ^ indexValue index
      | Jastore index => "\n\tastore " ^ indexValue index
      | Jistore_0     => "\n\tistore_0"
      | Jistore_1     => "\n\tistore_1" (*60*)
      | Jistore_2     => "\n\tistore_2"
      | Jistore_3     => "\n\tistore_3"
      | Jlstore_0     => "\n\tlstore_0"
      | Jlstore_1     => "\n\tlstore_1"
      | Jlstore_2     => "\n\tlstore_2"
      | Jlstore_3     => "\n\tlstore_3"
      | Jfstore_0     => "\n\tfstore_0"
      | Jfstore_1     => "\n\tfstore_1"
      | Jfstore_2     => "\n\tfstore_2"
      | Jfstore_3     => "\n\tfstore_3" (*70*)
      | Jdstore_0     => "\n\tdstore_0"
      | Jdstore_1     => "\n\tdstore_1"
      | Jdstore_2     => "\n\tdstore_2"
      | Jdstore_3     => "\n\tdstore_3"
      | Jastore_0     => "\n\tastore_0"
      | Jastore_1     => "\n\tastore_1"
      | Jastore_2     => "\n\tastore_2"
      | Jastore_3     => "\n\tastore_3"
      | Jiastore      => "\n\tiastore"
      | Jlastore      => "\n\tlastore"  (*80*)
      | Jfastore      => "\n\tfastore"
      | Jdastore      => "\n\tdastore"
      | Jaastore      => "\n\taastore"
      | Jbastore      => "\n\tbastore"
      | Jcastore      => "\n\tcastore"
      | Jsastore      => "\n\tsastore"
      | Jpop          => "\n\tpop"
      | Jpop2         => "\n\tpop2"
      | Jdup          => "\n\tdup"
      | Jdup_x1       => "\n\tdup_x1"   (*90*)
      | Jdup_x2       => "\n\tdup_x2"
      | Jdup2         => "\n\tdup2"
      | Jdup2_x1      => "\n\tdup2_x1"
      | Jdup2_x2      => "\n\tdup2_x2"
      | Jswap         => "\n\tswap"
      | Jiadd         => "\n\tiadd"
      | Jladd         => "\n\tladd"
      | Jfadd         => "\n\tfadd"
      | Jdadd         => "\n\tdadd"
      | Jisub         => "\n\tisub"     (*100*)
      | Jlsub         => "\n\tlsub"
      | Jfsub         => "\n\tfsub"
      | Jdsub         => "\n\tdsub"
      | Jimul         => "\n\timul"
      | Jlmul         => "\n\tlmul"
      | Jfmul         => "\n\tfmul"
      | Jdmul         => "\n\tdmul"
      | Jidiv         => "\n\tidiv"
      | Jldiv         => "\n\tldiv"
      | Jfdiv         => "\n\tfdiv"     (*110*)
      | Jddiv         => "\n\tddiv"
      | Jirem         => "\n\tirem"
      | Jlrem         => "\n\tlrem"
      | Jfrem         => "\n\tfrem"
      | Jdrem         => "\n\tdrem"
      | Jineg         => "\n\tineg"
      | Jlneg         => "\n\tlneg"
      | Jfneg         => "\n\tfneg"
      | Jdneg         => "\n\tdneg"
      | Jishl         => "\n\tishl"     (*120*)
      | Jlshl         => "\n\tlshl"
      | Jishr         => "\n\tishr"
      | Jlshr         => "\n\tlshr"
      | Jiushr        => "\n\tiushr"
      | Jlushr        => "\n\tlushr"
      | Jiand         => "\n\tiand"
      | Jland         => "\n\tland"
      | Jior          => "\n\tior"
      | Jlor          => "\n\tlor"
      | Jixor         => "\n\tixor"     (*130*)
      | Jlxor         => "\n\tlxor"
      | Jiinc { var   : Localvar.index,
		const : int} => "\n\tiinc " ^
		                indexValue var ^
				" " ^ Int.toString const
      | Ji2l   => "\n\ti2l"
      | Ji2f   => "\n\ti2f"
      | Ji2d   => "\n\ti2d"
      | Jl2i   => "\n\tl2i"
      | Jl2f   => "\n\tl2f"
      | Jl2d   => "\n\tl2d"
      | Jf2i   => "\n\tf2i"
      | Jf2l   => "\n\tf2l"             (*140*)
      | Jf2d   => "\n\tf2d"
      | Jd2i   => "\n\td2i"
      | Jd2l   => "\n\td2l"
      | Jd2f   => "\n\td2f"
      | Ji2b   => "\n\ti2b"
      | Ji2c   => "\n\ti2c"
      | Ji2s   => "\n\ti2s"
      | Jlcmp  => "\n\tlcmp"
      | Jfcmpl => "\n\tfcmpl"
      | Jfcmpg => "\n\tfcmpg"           (*150*)
      | Jdcmpl => "\n\tdcmpl"
      | Jdcmpg => "\n\tdcmpg"
      | Jifeq      label => "\n\tifeq "      ^ labelValue label 
      | Jifne      label => "\n\tifne "      ^ labelValue label
      | Jiflt      label => "\n\tiflt "      ^ labelValue label
      | Jifge      label => "\n\tifge "      ^ labelValue label
      | Jifgt      label => "\n\tifgt "      ^ labelValue label
      | Jifle      label => "\n\tifle "      ^ labelValue label
      | Jif_icmpeq label => "\n\tif_icmpeq " ^ labelValue label
      | Jif_icmpne label => "\n\tif_icmpne " ^ labelValue label  (*160*)
      | Jif_icmplt label => "\n\tif_icmplt " ^ labelValue label
      | Jif_icmpge label => "\n\tif_icmpge " ^ labelValue label
      | Jif_icmpgt label => "\n\tif_icmpgt " ^ labelValue label
      | Jif_icmple label => "\n\tif_icmple " ^ labelValue label
      | Jif_acmpeq label => "\n\tif_acmeqp " ^ labelValue label
      | Jif_acmpne label => "\n\tif_acmpne " ^ labelValue label
      | Jgoto      label => "\n\tgoto "      ^ labelValue label
      | Jjsr       label => "\n\tjsr "       ^ labelValue label
      | Jret       index => "\n\tret "       ^ indexValue index
      | Jtableswitch { default : Label.label
                     , offset  : Int32.int   (* key of first target *)
		     , targets : Label.label Vector.vector } =>
		       let fun add_nt s =  "\n\t\t" ^ s
		           val tars  = Vector.foldr op^ "" (Vector.map (add_nt o labelValue) targets)  (*170*)
		           val instr = "\n\ttableswitch " ^ (Int.toString o Int32.toInt) offset
			   val def   = "\n\t\tdefault : " ^ labelValue default
		       in instr ^ tars ^ def
		       end
      | Jlookupswitch { default
                      , cases } => let fun ppKeyLab (key, lab) = "\n\t\t<unimplemented int> : " ^ labelValue lab
        		               val instr = "\n\tlookupswitch"
				       val table = concat (map ppKeyLab cases)
				       val def   = "\n\t\tdefault : " ^ labelValue default
				   in  instr ^ table ^ def
				   end
      | Jireturn => " \n\tireturn"
      | Jlreturn => " \n\tlreturn"
      | Jfreturn => " \n\tfreturn"
      | Jdreturn => " \n\tdreturn"
      | Jareturn => " \n\tareturn"
      | Jreturn  => " \n\treturn"
      | Jgetstatic field_ref => "\n\tgetstatic " ^
                                className (#class field_ref) ^
			        "/" ^ (#name field_ref) ^
			        " " ^ ppType (#ty field_ref)
      | Jputstatic field_ref => "\n\tputstatic " ^
                                className (#class field_ref) ^
			        "/" ^ (#name field_ref) ^
			        " " ^ ppType (#ty field_ref)
      | Jgetfield  field_ref => "\n\tgetfield " ^
                                className (#class field_ref) ^
			        "/" ^ (#name field_ref) ^
			        " " ^ ppType (#ty field_ref)  (*180*)
      | Jputfield  field_ref => "\n\tputfield " ^
                                className (#class field_ref) ^
			        "/" ^ (#name field_ref) ^
			        " " ^ ppType (#ty field_ref)
      | Jinvokevirtual   method_ref => let val msig = (#msig method_ref)
                                       in  "\n\tinvokevirtual " ^
                                           className (#class method_ref) ^
                                           "/" ^ (#name method_ref) ^
				           "(" ^ concat (map ppType (#1 msig)) ^ ")" ^
					   ppType_option (#2 msig)
				       end
      | Jinvokespecial   method_ref => let val msig = (#msig method_ref)
                                       in  "\n\tinvokespecial " ^
                                           className (#class method_ref) ^
                                           "/" ^ (#name method_ref) ^
				           "(" ^ concat (map ppType (#1 msig)) ^ ")" ^
					   ppType_option (#2 msig)
				       end
      | Jinvokestatic    method_ref => let val msig = (#msig method_ref)
                                       in  "\n\tinvokestatic " ^
                                           className (#class method_ref) ^
                                           "/" ^ (#name method_ref) ^
				           "(" ^ concat (map ppType (#1 msig)) ^ ")" ^
					   ppType_option (#2 msig)
				       end
      | Jinvokeinterface method_ref => let val msig = (#msig method_ref)
                                       in  "\n\tinvokeinterface " ^
                                           className (#class method_ref) ^
                                           "/" ^ (#name method_ref) ^
				           "(" ^ concat (map ppType (#1 msig)) ^ ")" ^
					   ppType_option (#2 msig)
				       end
      | Jnew       jclass    => "\n\tnew " ^ className jclass
      | Jnewarray  elem      => "\n\tnewarray " ^ ppType elem
      | Janewarray class_ref => "\n\tanewarray " ^
                       ( case class_ref of
		           CLASS c => className c
			 | ARRAY a => "[" ^ ppType a
		       )
      | Jarraylength => "\n\tarray_length"  (*190*)
      | Jathrow      => "\n\tathrow" 
      | Jcheckcast  class_ref => "\n\tcheckcast " ^
                                 ( case class_ref of
				     CLASS c => className c
				   | ARRAY t => ppType t
			 	 )
      | Jinstanceof class_ref => "\n\tinstanceof " ^
                                 ( case class_ref of
				     CLASS c => className c
				   | ARRAY t => ppType t
				 )
      | Jmonitorenter    => "\n\tmonitorenter"
      | Jmonitorexit     => "\n\tmonitorexit"
      | Jwide  jvm_instr => "\n\twide " ^ pp_instr jvm_instr
      | Jmultianewarray { elem : jtype
		        , dim  : int } => "\n\tmultianewarray " ^
			                  ppType elem ^ " " ^
					  Int.toString dim
      | Jifnull    label => "\n\tifnull "    ^ labelValue label
      | Jifnonnull label => "\n\tifnonnull " ^ labelValue label
      | Jgoto_w    label => "\n\tgoto_w "    ^ labelValue label  (*200*)
      | Jjsr_w     label => "\n\tjsr_w "     ^ labelValue label      
      | _                => "\n\n\n!!! UNKNOWN BYTECODE !!!\n\n"

   (* pretty_print the bytecode instructions *)
   fun pp_bytecode instr_list = concat (map pp_instr instr_list)
   
   (* pretty_print the .catch statements *)
   fun pp_hdls pool li =
   let fun extract_info (hdl:exn_hdl) =
       let val start = (Int.toString o Word16.toInt) (#start hdl)
           val stop  = (Int.toString o Word16.toInt) (#stop hdl)
           val entry = (Int.toString o Word16.toInt) (#entry hdl)
       in  case (#catch hdl) of
             (SOME idx) => "\n.catch " ^ Constpool.extract_class pool idx ^
	                   " from " ^ start ^
		           " to " ^ stop ^
		           " using " ^ entry
           |  NONE      => "??? no catch type ???"
       end
   in  concat (map extract_info li)
   end

   (* pretty_print the .throws statements *)
   fun pp_exns pool li =
   let fun extract_index idx = "\n.throws " ^ Constpool.extract_class pool idx
   in  concat (map (extract_index) li)
   end
in 
   fun pp_code pool [] = ""
     | pp_code pool (x::xs) =
       case x of
         CODE c => let val stack  = (Int.toString o Word16.toInt) (#stack c)
                       val locals = (Int.toString o Word16.toInt) (#locals c)
		       val code   = pp_bytecode (Parser.parseByteCode (#code c) pool)
		   in "\n.limit stack " ^ stack ^
		      "\n.limit locals " ^ locals ^
		      code ^
		      pp_hdls pool (#hdls c) ^  (* .catch  *)
		      pp_code pool xs           (* find .throws *)
		   end
       | EXNS e => pp_exns pool (#exns e) ^ pp_code pool xs
       | _      => pp_code pool xs
end

(* pretty_print a .method statement *)
fun pp_method pool (mem:member) = 
let val flags = Word16.toInt (#flags mem)
    val mods  = pp_modifiers flags "method"
    val name  = Constpool.extract_name pool (#name mem)
    val desc  = Constpool.extract_name pool (#desc mem) ^ " "
    val code  = pp_code pool (#attrs mem)
in  "\n\n.method " ^ mods ^ name ^ desc ^ code ^ "\n.end method"
end

(* MAIN FUNCTION *)
fun pretty_print (cf:class_file) =
let val pool = #pool cf
    val pp_super   = case (#super cf) of
                       NONE     => ""
                     | SOME idx => "\n.super " ^
		                   Constpool.extract_class pool idx ^ "\n"
    val mods = Word16.toInt (#flags cf)
    val pp_class   = "\n.class " ^ pp_modifiers mods "class" ^
                     Constpool.extract_class pool (#this cf)
    val pp_header  = pp_srcfile (#attrs cf) pool ^
                     pp_class ^
		     pp_super
    val pp_ifc     = (fn x => "\n.implements " ^ x) o
		     (Constpool.extract_class pool)
    val pp_impl    = concat ( map pp_ifc (#ifcs cf) )
    val pp_fields  = concat ( map (pp_field pool)
                                  (#fields cf)
			    )
    val pp_methods = concat ( map (pp_method pool)
                                  (#methods cf)
		            )
in  print (pp_header ^ pp_impl ^ pp_fields ^ pp_methods)
end

end
