(*                                                                      *)
(* MLJVM - a Java Virtual Machine implemented in Standard ML            *)
(*                                                                      *)
(* http://homepages.inf.ed.ac.uk/stg/research/projects/MLJVM/           *)
(*                                                                      *)
(* Copyright (C) 2003 Stephen Gilmore and Laura Korte, The              *)
(* University of Edinburgh.                                             *)
(*                                                                      *)
(* This program is free software; you can redistribute it and/or        *)
(* modify it under the terms of the GNU General Public License          *)
(* as published by the Free Software Foundation; either version 2       *)
(* of the License, or (at your option) any later version.               *)
(*                                                                      *)
(* This program is distributed in the hope that it will be useful,      *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of       *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *)
(* GNU General Public License for more details.                         *)
(*                                                                      *)
(* You should have received a copy of the GNU General Public License    *)
(* along with this program; if not, write to the Free Software          *)
(* Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *)
(* MA 02111-1307, USA.                                                  *)
(*                                                                      *)

structure Parser :> Parser =
struct

exception Error of string;
exception UnsupportedAttribute of string;
exception AttributeFormatException;
exception UnsupportedCPitem of string;
exception CannotDo of string
exception KeyNotFoundError;
exception UnknownInstructionError of int;

open MyClassFile

(*** start debug-info ***)
val debug_string = ref "" : string ref
fun do_debug 0 ss = ()
  | do_debug 1 ss = ()
  | do_debug 2 ss = ()
  | do_debug 3 ss = ()
  | do_debug n ss =  ((print ss);(TextIO.flushOut TextIO.stdOut))
(**** end debug-info ****)


(*********************************************************************************************)
(*************************************** auxilary functions **********************************)
(*********************************************************************************************)


(* add_idx_option : int -> index option *)
fun add_idx_option 0 = NONE
  | add_idx_option n = SOME (Constpool.makeIndex n)

(* apply <parser> to <vector> <number> times *)
(* NB. only for parsers which yield strings  *)
fun <*> parser (result,vector_index,number) vector =
        if number > 0
        then let val (s,new_vi) = parser vector vector_index
	     in  <*> parser (result ^ s, new_vi, number-1) vector
	     end
	else (result,vector_index)

(* apply <parser> to <vector> <number> times *)
fun <**> parser (result,vector_index,number) vector =
let fun <++> parser (result,vector_index,number) vector =
        if number > 0
        then let val (s,new_vi) = parser vector vector_index
	     in  <++> parser (s :: result, new_vi, number-1) vector
	     end
	else (result,vector_index)
    val (result, vector_index) = <++> parser (result,vector_index,number) vector
in  (rev result, vector_index)
end

(* apply <parser> to <vector> until <size> is reached *)
fun <***> parser (result,vector_index,size) vector =
let fun <+++> parser (result,vector_index,number) vector =
        if vector_index < size
        then let val (s,new_vi) = parser vector vector_index
	     in  <+++> parser (s :: result, new_vi, size) vector
	     end
	else (result,vector_index)
    val (result, vector_index) = <+++> parser (result,vector_index,size) vector
in  (rev result, vector_index)
end

local
   fun i2i_i (tokens,n) = (Word8.toInt o Word8Vector.sub) (tokens, n)
in
   (* byte2int functions *)
   fun parseU1_i v n =
   let val new_index = n+1
   in  (i2i_i (v,n), new_index)
   end

   fun parseU2_i v n =
   let val new_index = n+2
   in  ( (i2i_i (v,n))*16*16 + (i2i_i (v,n+1))
      , new_index )
   end

   fun parseU2_offset v n =
   let val new_index = n+2
       val int8 = Word.fromInt 8
       val aux_word = 0wx7FFF0000
       val aux_int = 32768 * 2
       val word1 = Word8Vector.sub (v,n)
       val word2 = Word8Vector.sub (v,n+1)
       val return_int_aux = Word.toIntX ( Word.orb ( Word.orb ( Word.<< (Word8.toLargeWord word1,int8)
                                                              , Word8.toLargeWord word2 )
                                                   , aux_word ))
       val return_int' = Word.toIntX ( Word.orb ( Word.<< (Word8.toLargeWord word1,int8)
                                                           , Word8.toLargeWord word2 ))
       val return_int = if return_int' < 32768
                        then return_int'
			else return_int' - aux_int
   in  ( return_int
       , new_index )
   end 

   fun parseU4_i v n =
   let val new_index = n+4
       val p2 = 16*16
       val p4 = p2*p2
       val p6 = p4*p2
   in  (   ((i2i_i (v,n  ))*p6 )
         + ((i2i_i (v,n+1))*p4 )
         + ((i2i_i (v,n+2))*p2 )
         + ((i2i_i (v,n+3))    )
       , new_index  )
   end

   (* byte2idx functions *)
   fun parseU1_idx v n =
   let val new_index = n+1
   in  ( Constpool.makeIndex (i2i_i (v,n))
       , new_index )
   end

   fun parseU2_idx v n =
   let val new_index = n+2
   in  ( Constpool.makeIndex ((i2i_i (v,n))*16*16 + (i2i_i (v,n+1)))
       , new_index )
   end
end

(* byte2word functions *)
local
   fun stripSOME (SOME x) = x
     | stripSOME NONE = raise CannotDo "stripSOME"
in
   fun parseU2_w16 v n =
   let val u2_vector = Word8Vector.extract (v, n, SOME 2)
   in  ( stripSOME (Word16.fromBytes u2_vector)
       , n+2 )
   end

   fun parseU4_w32 v n =
   let val u4_vector = Word8Vector.extract (v, n, SOME 4)
   in  ( stripSOME (Word32.fromBytes u4_vector)
       , n+4 )
   end

   fun parseU4_i32 v n =
   let val u4_vector = Word8Vector.extract (v, n, SOME 4)
   in  ( stripSOME (Int32.fromBytes u4_vector)
       , n+4 )
   end

   fun parseU8_i64 v n =
   let val u8_vector = Word8Vector.extract (v, n, SOME 8)
   in  ( stripSOME (Int64.fromBytes u8_vector)
       , n+8 )
   end
end

(* byte2byte functions *)
fun parseU4_w v n =
let val u4_vector = Word8Vector.extract (v, n, SOME 4)
in  ( u4_vector
    , n+4 )
end

fun parseU8_w v n =
let val u8_vector = Word8Vector.extract (v, n, SOME 8)
in  ( u8_vector
    , n+8 )
end

(*********************************************************************************************)
(******************************** end of auxilary functions **********************************)
(*********************************************************************************************)

(* constantpool-item-parsers *)
local
   open Constpool
in 
   fun parseUtf8 v n =
      let val tag     = "Utf8"
          val (len,i) = parseU2_i v n
          fun intWord8toS (tokens,n) = (String.str o Char.chr o Word8.toInt o Word8Vector.sub) (tokens, n)
          fun parseChar v m  = (intWord8toS (v,m),m+1)
          val (s,i)   = <*> parseChar ("",i,len) v
   in  ( CPutf8 s
       , i
       , 1 )
   end

   fun parseMethodref v n =
   let val tag = "MethodRef"
       val (class_index        ,i) = parseU2_idx v n
       val (name_and_type_index,i) = parseU2_idx v i
   in  ( CPmethodref { class    = class_index
                     , nameType = name_and_type_index }
       , i
       , 1 )
   end

   fun parseFieldref v n =
   let val tag = "FieldRef"
       val (class_index        ,i) = parseU2_idx v n
       val (name_and_type_index,i) = parseU2_idx v i
   in  ( CPfieldref { class    = class_index
                    , nameType = name_and_type_index }
       , i
       , 1 )
   end

   fun parseInterfaceMethodref v n =
   let val tag = "InterfaceMethodRef"
       val (class_index        ,i) = parseU2_idx v n
       val (name_and_type_index,i) = parseU2_idx v i
   in  ( CPimethodref { class    = class_index
                      , nameType = name_and_type_index }
       , i
       , 1 )
   end

   fun parseNameAndType v n =
   let val tag = "NameAndType"
       val (name_index      ,i) = parseU2_idx v n
       val (descriptor_index,i) = parseU2_idx v i
   in  ( CPnametype { name = name_index
                    , desc = descriptor_index }
       , i
       , 1 )
   end

   fun parseClass v n =
   let val tag = "Class"
       val (name_index,i) = parseU2_idx v n
   in  ( CPclass name_index
       , i
       , 1 )
   end


   fun parseString v n =
   let val tag = "String"
       val (name_index,i) = parseU2_idx v n
   in  ( CPstring name_index
       , i
       , 1 )
   end


   fun parseInteger v n = 
   let val tag = "Integer"
       val (int_vector,i) = parseU4_i32 v n
   in  ( CPint int_vector
       , i
       , 1 )
   end

   fun parseFloat v n = 
   let val tag = "Float"
       val (float_vector,i) = parseU4_w v n
   in  ( CPfloat float_vector
       , i
       , 1 )
   end

   (* used to have "high_bytes" & "low_bytes"*)
   fun parseLong v n =
   let val tag = "Long"
       val (long_vector,i) = parseU8_i64 v n
   in  ( CPlong long_vector
       , i
       , 2 )
   end

   fun parseDouble v n = 
   let val tag = "Double"
       val (double_vector,i) = parseU8_w v n
   in  ( CPdouble double_vector
       , i
       , 2 )
   end
end

fun parseCP v n size = 
let (* <cp_array> use "size", because 0 is not a valid index *)
    val cp_array = Array.array(size,Constpool.CPunused)
    fun parseItem (integer,m) = 
        case integer of 1  => parseUtf8 v m
                      | 3  => parseInteger v m
                      | 4  => parseFloat v m
                      | 5  => parseLong v m
                      | 6  => parseDouble v m
                      | 7  => parseClass v m
                      | 8  => parseString v m
                      | 9  => parseFieldref v m
                      | 10 => parseMethodref v m
                      | 11 => parseInterfaceMethodref v m
                      | 12 => parseNameAndType v m
                      | _  => raise UnsupportedCPitem (Int.toString integer)
    fun parseTag m = parseU1_i v m
    fun parseItems (u,m,z) = if z < size
                             then let val (j,y,delta_index) = parseItem (parseTag m)
			          in parseItems ( Array.update(cp_array, z, j)
						, y
						, z+delta_index )
				  end
			       else ((),m)
    val (_, i) = parseItems ((),n,1)
in  ( cp_array, i )
end

fun parseInterfaces v n size = <**> parseU2_idx ([],n,size) v

fun parseExceptionTable v n size = 
let fun parseItem v m = let val (start_pc  ,i) = parseU2_w16 v m
                            val (end_pc    ,i) = parseU2_w16 v i
	  		    val (handler_pc,i) = parseU2_w16 v i
                            val (catch_type,i) = parseU2_i v i
			    val my_catch = add_idx_option catch_type
		        in  ( { start = start_pc
			      , stop  = end_pc
			      , entry = handler_pc
			      , catch = my_catch } : exn_hdl
			    , i )
			end
in  <**> parseItem ([],n,size) v
end

local
   open Constpool
in
   (* parseAttributes has mut.rec. with parseInfo *)
   fun parseAttributes v n size pool =
   let fun parseItem w m = let val (attribute_name_index,j) = parseU2_i w m
			       val cp_item = Array.sub (pool,attribute_name_index)
			       val (attribute_length,j) = parseU4_i w j
			       val (attribute,j) = parseInfo w
			                                     j
							     attribute_length
							     pool
							     (cp_item, attribute_name_index)
                           in  ( attribute, j )
		           end
   in  <**> parseItem ([],n,size) v
   end

   (* parseAttributes has mut.rec. with parseInfo *)
   and parseInfo v n size pool (cp_item,attribute_name_index) =
   let val empty_vector = Word8Vector.fromList []
       fun parseLineNr v j = let val (start_pc,i) = parseU2_w16 v j
                                 val (line_number,i) = parseU2_w16 v i
		             in  ( { start = start_pc
			           , line  = line_number } : line_number_info
			         , i )
		             end
       fun parseLineNumberTable m table_length = <**> parseLineNr ([],m,table_length) v
   in  case cp_item of
         CPutf8 "Code" => let val (max_stack,i) = parseU2_w16 v n
       			      val (max_locals,i) = parseU2_w16 v i
			      val (code_length,i) = parseU4_i v i
			      val code_length_s = Int.toString code_length
			      val code_vector = Word8Vector.extract (v, i, SOME code_length)
			      val i = i+code_length
			      val (exception_table_length,i) = parseU2_i v i
			      val exception_table_length_s = Int.toString exception_table_length
                              val (exception_table,i) = parseExceptionTable v i exception_table_length
			      val (attributes_count,i) = parseU2_i v i
			      val attributes_count_s = Int.toString attributes_count
			      val (attributes,i) = parseAttributes v i attributes_count pool
		          in  ( CODE { attr   = Constpool.makeIndex attribute_name_index
			             , stack  = max_stack
			             , locals = max_locals
			             , code   = code_vector
			             , hdls   = exception_table
			             , attrs  = attributes }
	  	              , i )
		          end 
       | CPutf8 "SourceFile" => let val (index,i) = parseU2_i v n
			        in  ( SRCFILE { attr = Constpool.makeIndex attribute_name_index
	 	 		              , file = Constpool.makeIndex index }
			            , i )
			        end
       | CPutf8 "LineNumberTable" => let val (line_number_table_length,i) = parseU2_i v n
				         val (line_number_table,i) = parseLineNumberTable i line_number_table_length
			             in  ( LINENUM { attr = Constpool.makeIndex attribute_name_index
				  	           , lines = line_number_table }
				         , i )
			             end
       | CPutf8 "ConstantValue" => let val (constant_value_index,i) = parseU2_i v n
			           in  ( CONSTVAL { attr  = Constpool.makeIndex attribute_name_index
					          , value = Constpool.makeIndex constant_value_index }
				       , i )
			           end
       | CPutf8 "Exceptions" => let val (number_of_exceptions,i) = parseU2_i v n
       			            val (exception_index_table,i) = <**> parseU2_idx ([],i,number_of_exceptions) v
			        in  ( EXNS { attr = Constpool.makeIndex attribute_name_index
				           , exns = exception_index_table }
			            , i )
			        end
       | CPutf8 "InnerClasses" => let val (number_of_classes,i) = parseU2_i v n
	                              fun parseItem w m = let val (inner_info_index,i) = parseU2_i w m
				                              val (outer_info_index,i) = parseU2_i w i
                                                              val (inner_name_index,i) = parseU2_i w i
                                                              val (inner_flags     ,i) = parseU2_w16 w i
					  	          in  ( { inner_info  = add_idx_option inner_info_index
							        , outer_info  = add_idx_option inner_info_index
							        , inner_name  = add_idx_option inner_name_index
							        , inner_flags = inner_flags } : inner_class_info 
							      , i )
						          end
 				      val (classes, i) = <**> parseItem ([],i,number_of_classes) v
			          in  ( INNER { attr = Constpool.makeIndex attribute_name_index
					      , classes = classes }
				      , i )
			          end
       | CPutf8 "Synthetic" => if size = 0
	                       then ( ATTR { attr = Constpool.makeIndex attribute_name_index
				           , info = empty_vector }
			  	    , n )
			       else raise AttributeFormatException
       | CPutf8 "LocalVariableTable" => let val (table_length,i) = parseU2_i v n  (*untested*)
	                                    fun parseItem v m = let val (start_pc  ,j) = parseU2_w16 v m
					                            val (length    ,j) = parseU2_w16 v j
							            val (name_index,j) = parseU2_idx v j
							            val (desc_index,j) = parseU2_idx v j
							            val (index,j) = parseU2_w16 v j
							        in  ( { start  = start_pc
								       , length = length
								       , name   = name_index
								       , desc   = desc_index
								       , index  = index } : local_var_info 
							             , j )
							        end
				             val (my_vars,i) = <**> parseItem ([],i,table_length) v
				        in  ( LOCALVAR { attr = Constpool.makeIndex attribute_name_index
					               , vars = my_vars }
				            , i )
				        end
       | CPutf8 "Deprecated" => if size = 0
	                        then ( ATTR { attr = Constpool.makeIndex attribute_name_index
				            , info = empty_vector }
				     , n )
			        else raise AttributeFormatException
       | CPutf8 _            => ( ATTR { attr = Constpool.makeIndex attribute_name_index
				            , info = empty_vector }
				     , n + size )
(*       | CPutf8 identifier => raise UnsupportedAttribute identifier*)
       | _                 => raise Error "uftf8 expected in <parseInfo>"
   end
end

fun parseFields v n size pool =
let fun parseItem v m = let val (access_flags,j) = parseU2_w16 v m
                            val (name_index,j) = parseU2_idx v j
 			    val (descriptor_index,j) = parseU2_idx v j
 			    val (attributes_count,j) = parseU2_i v j
 			    val attributes_count_s = Int.toString attributes_count
 			    val (attributes,j) = parseAttributes v j attributes_count pool
 			    val my_member = { flags = access_flags
			                    , name  = name_index
					    , desc  = descriptor_index
					    , attrs = attributes } : member
		        in  ( my_member
			    , j )
		        end
in  <**> parseItem ([],n,size) v
end

fun parse infile = 
let val is = BinIO.openIn infile
    val tokens = BinIO.inputAll is
    val len = Word8Vector.length tokens

    val empty_hash_table = Polyhash.mkPolyTable (0,KeyNotFoundError)

    val (magic,i) = parseU4_w32 tokens 0
    val (minor,i) = parseU2_w16 tokens i
    val (major,i) = parseU2_w16 tokens i

    val (cp_count,i) = parseU2_i tokens i
    val cp_count_s = Int.toString cp_count
    val (pool_array,i) = parseCP tokens i cp_count

    val (access_flags,i) = parseU2_w16 tokens i
    val (this_class,i) = parseU2_i tokens i
    val this_class_s = Int.toString this_class
    val (super_class,i) = parseU2_i tokens i     (* = 0 or index *)
    val super_class_s = Int.toString super_class
    val super_class_option = add_idx_option super_class

    val (interfaces_count,i) = parseU2_i tokens i
    val interfaces_count_s = Int.toString interfaces_count
    val (interfaces,i) = parseInterfaces tokens i interfaces_count

    val (fields_count,i) = parseU2_i tokens i
    val fields_count_s = Int.toString fields_count
    val (fields,i) = parseFields tokens i fields_count pool_array

    val (methods_count,i) = parseU2_i tokens i
    val methods_count_s = Int.toString methods_count
    val (methods,i) = parseFields tokens i methods_count pool_array

    val (attributes_count,i) = parseU2_i tokens i
    val attributes_count_s = Int.toString attributes_count
    val (attributes,i) = parseAttributes tokens i attributes_count pool_array
    val my_class_file = { magic   = magic
                        , minor   = minor
			, major   = major
			, pool    = Constpool.createPool ( ref pool_array
			                                 , empty_hash_table
				                         , ref (Array.length pool_array) )
			, flags   = access_flags
			, this    = Constpool.makeIndex this_class
			, super   = super_class_option
			, ifcs    = interfaces
			, fields  = fields
			, methods = methods
			, attrs   = attributes } : class_file 
    val _ =  BinIO.closeIn is
in  if i = len
    then my_class_file
    else raise (Error "File too short")
end















fun parseType_aux charlist =
let val h = hd charlist
    val t = tl charlist
    fun scan (#";"::cs,li) = (cs, #";" :: li)
      | scan (c::cs,li)    = scan (cs,c :: li)
      | scan ([],li)       = ([],li)
in  case h of #"B" => ( Jvmtype.Tbyte, t )
            | #"C" => ( Jvmtype.Tchar, t )
            | #"D" => ( Jvmtype.Tdouble, t )
            | #"F" => ( Jvmtype.Tfloat, t )
            | #"I" => ( Jvmtype.Tint, t )
            | #"J" => ( Jvmtype.Tlong, t )
            | #"S" => ( Jvmtype.Tshort, t )
            | #"Z" => ( Jvmtype.Tboolean, t )
            | #"L" => ( fn (junk, result) => (Jvmtype.Tclass (JvmType.class {pkgs=[],name=implode (rev result)}) , junk)
                      )    (scan (t,[]))
            | #"[" => ( fn (result, junk) => ( Jvmtype.Tarray result, junk )
                      )    (parseType_aux t) 
            | _    => raise CannotDo "parseType"
end

fun parseType s =
let val (result,junk) = (parseType_aux o explode) s
in  if junk = []
    then result
    else raise Error "while parsing type"
end


   (*there must be a better solution than duplicating this function!*)
fun parseType_option charlist =
let val h = hd charlist
    val t = tl charlist
    fun scan (#";"::cs,li) = (cs, #";" :: li)
      | scan (c::cs,li)    = scan (cs,c :: li)
      | scan ([],li)       = ([],li)
in  case h of #"B" => ( SOME Jvmtype.Tbyte, t )
            | #"C" => ( SOME Jvmtype.Tchar, t )
            | #"D" => ( SOME Jvmtype.Tdouble, t )
            | #"F" => ( SOME Jvmtype.Tfloat, t )
            | #"I" => ( SOME Jvmtype.Tint, t )
            | #"J" => ( SOME Jvmtype.Tlong, t )
            | #"S" => ( SOME Jvmtype.Tshort, t )
            | #"Z" => ( SOME Jvmtype.Tboolean, t )
	    | #"L" => ( fn (junk, result) =>
	                ( SOME ( Jvmtype.Tclass (JvmType.class { pkgs = []
	                                                       , name = implode (rev result) } ) )
			, junk )
	              ) (scan (t,[]))
            | #"[" => ( fn (result, junk) => ( SOME (Jvmtype.Tarray result), junk )
	              )    (parseType_aux t) 
            | #"V" => ( NONE, t ) 
            | _    => raise CannotDo "parseType_option"
end
   
fun parseTypes s =
let val clist = tl (explode s)
    fun somefunc (res,#")"::cs)    = (res,cs)
      | somefunc (res,c::cs)       = ( fn (result, junk) => somefunc (result::res,junk)
                                     ) (parseType_aux (c::cs))
      | somefunc _                 = raise CannotDo "parseTypes"
    val (result_list, rest) = somefunc ([],clist)
in  ( rev result_list
    , (#1 o parseType_option) rest )
end

local open Bytecode Localvar   
   fun getMethod_ref pool idx =
   let val fld = Constpool.lookup pool idx
       val (cl_idx,naty_idx) = case fld of
                                 Constpool.CPmethodref {class, nameType} => (class, nameType)
			       | n => raise Error "MethodExpectedError"
       val naty    = Constpool.lookup pool naty_idx
       val cl_name = Constpool.extract_class pool cl_idx
       val cl      = Jvmtype.class {name=cl_name, pkgs=[]}
       val (na_idx,ty_idx) = case naty of
                               Constpool.CPnametype {name,desc} => (name, desc)
			     | n => raise Error "NameTypeExpectedError"
       val na_val      = Constpool.extract_name pool na_idx
       val type_string = Constpool.extract_name pool ty_idx
       val ty_val      = parseTypes type_string
   in  { class = cl
       , name  = na_val
       , msig = ty_val }
   end

   fun getField_ref pool idx =
   let val fld = Constpool.lookup pool idx
       val (cl_idx,naty_idx) = case fld of
                                 Constpool.CPfieldref {class, nameType} => (class, nameType)
			       | n => raise Error "FieldExpectedError"
       val naty = Constpool.lookup pool naty_idx
       val cl_name = Constpool.extract_class pool cl_idx
       val cl = Jvmtype.class {name=cl_name, pkgs=[]}
       val (na_idx,ty_idx) = case naty of Constpool.CPnametype {name,desc} => (name, desc)
			                | n            => raise Error "NameTypeExpectedError"
       val na_val = Constpool.extract_name pool na_idx
       val ty_val = parseType (Constpool.extract_name pool ty_idx)
   in  { class = cl
       , name  = na_val
       , ty    = ty_val }
   end

   fun parseClass_ref pool idx =
   let val name = Constpool.extract_class pool idx
       val char_list = explode name
       fun choose clist = case (hd clist) of
                            #"[" => ARRAY ((#1 o parseType_aux) (tl clist))
                          | x    => CLASS (Jvmtype.class {name=name,pkgs=[]})
   in  choose char_list
   end
				        
in 
   fun parseByteCode v pool =
   let val size = Word8Vector.length v
       fun parseCode v i = let val (instr,i) = parseU1_i v i
			   in case instr of
			          0 => ( Jnop        , i )
			      |   1 => ( Jaconst_null, i )
			      |   2 => ( Jiconst_m1  , i )
			      |   3 => ( Jiconst_0   , i )
			      |   4 => ( Jiconst_1   , i )
			      |   5 => ( Jiconst_2   , i )
			      |   6 => ( Jiconst_3   , i )
			      |   7 => ( Jiconst_4   , i )
			      |   8 => ( Jiconst_5   , i )
			      |   9 => ( Jlconst_0   , i )
			      |  10 => ( Jlconst_1   , i )
			      |  11 => ( Jfconst_0   , i )
			      |  12 => ( Jfconst_1   , i )
			      |  13 => ( Jfconst_2   , i )
			      |  14 => ( Jdconst_0   , i )
			      |  15 => ( Jdconst_1   , i )
			      |  16 => ( fn (byte,j) => ( Jbipush byte, j )
				       ) (parseU1_i v i)
			      |  17 => ( fn (bytes2,j) => ( Jsipush bytes2, j )
				       ) (parseU2_i v i)
			      |  18 => let val (idx, i) = (parseU1_i v i)
                                           val cp_item = Constpool.lookup pool (Constpool.makeIndex idx)
                                       in  case cp_item of
					     Constpool.CPint integer    => (Jldc (Cint integer), i)
					   | Constpool.CPfloat float    =>
					     let val r_opt = (Real32.fromBytes float)
					     in  case r_opt of
					           SOME r => (Jldc (Cfloat r), i)
						 | NONE   => raise CannotDo "pCode 18a"
					     end
					   | Constpool.CPstring str_ref =>
					     let val cp_item = (Constpool.lookup pool str_ref)
					     in  case cp_item of
					           Constpool.CPutf8 s => (Jldc (Cstring s), i)
						 | _                  => raise CannotDo "pCode 18b"
					     end
					   | _                          => raise CannotDo "pCode 18c"
				       end
			      |  19 => let val (idx, i) = (parseU2_idx v i)
                                           val cp_item = Constpool.lookup pool idx
                                       in  case cp_item of
					     Constpool.CPint integer    => (Jldc_w (Cint integer), i)
					   | Constpool.CPfloat float    =>
					     let val r_opt = (Real32.fromBytes float)
					     in  case r_opt of
					           SOME r => (Jldc_w (Cfloat r), i)
						 | NONE   => raise CannotDo "pCode 18a"
					     end
					   | Constpool.CPstring str_ref =>
					     let val cp_item = (Constpool.lookup pool str_ref)
					     in  case cp_item of
					           Constpool.CPutf8 s => (Jldc_w (Cstring s), i)
						 | _                  => raise CannotDo "pCode 18b"
					     end
					   | _                          => raise CannotDo "pCode 18c"
				       end
			      |  20 => let val (idx, i) = (parseU2_idx v i)
                                           val cp_item = Constpool.lookup pool idx
                                       in  case cp_item of
					     Constpool.CPdouble double =>
					     let val d_opt = (Real64.fromBytes double)
					     in  case d_opt of
					           SOME d => (Jldc2_w (Cdouble d), i)
						 | NONE   => raise CannotDo "pCode 20a"
					     end
					   | Constpool.CPlong   long   => (Jldc2_w (Clong long), i)
					   | _                         => raise CannotDo "pCode 20c"
				       end
			      |  21 => ( fn (byte,j) => ( Jiload (fromInt byte), j )
				       ) (parseU1_i v i)
			      |  22 => ( fn (byte,j) => ( Jlload (fromInt byte), j )
				       ) (parseU1_i v i)
			      |  23 => ( fn (byte,j) => ( Jfload (fromInt byte), j )
				       ) (parseU1_i v i)
			      |  24 => ( fn (byte,j) => ( Jdload (fromInt byte), j )
				       ) (parseU1_i v i)
			      |  25 => ( fn (byte,j) => ( Jaload (fromInt byte), j )
				       ) (parseU1_i v i)
			      |  26 => ( Jiload_0 , i )
			      |  27 => ( Jiload_1 , i )
			      |  28 => ( Jiload_2 , i )
			      |  29 => ( Jiload_3 , i )
			      |  30 => ( Jlload_0 , i )
			      |  31 => ( Jlload_1 , i )
			      |  32 => ( Jlload_2 , i )
			      |  33 => ( Jlload_3 , i )
			      |  34 => ( Jfload_0 , i )
			      |  35 => ( Jfload_1 , i )
			      |  36 => ( Jfload_2 , i )
			      |  37 => ( Jfload_3 , i )
			      |  38 => ( Jdload_0 , i )
			      |  39 => ( Jdload_1 , i )
			      |  40 => ( Jdload_2 , i )
			      |  41 => ( Jdload_3 , i )
			      |  42 => ( Jaload_0 , i ) 
			      |  43 => ( Jaload_1 , i )
			      |  44 => ( Jaload_2 , i )
			      |  45 => ( Jaload_3 , i )
			      |  46 => ( Jiaload  , i )
			      |  47 => ( Jlaload  , i )
			      |  48 => ( Jfaload  , i )
			      |  49 => ( Jdaload  , i )
			      |  50 => ( Jaaload  , i )
			      |  51 => ( Jbaload  , i )
			      |  52 => ( Jcaload  , i )
			      |  53 => ( Jsaload  , i )
			      |  54 => ( fn (byte,j) => ( Jistore (fromInt byte), j )
				       ) (parseU1_i v i)
			      |  55 => ( fn (byte,j) => ( Jlstore (fromInt byte), j )
				       ) (parseU1_i v i)
			      |  56 => ( fn (byte,j) => ( Jfstore (fromInt byte), j )
				       ) (parseU1_i v i)
			      |  57 => ( fn (byte,j) => ( Jdstore (fromInt byte), j )
				       ) (parseU1_i v i)
			      |  58 => ( fn (byte,j) => ( Jastore (fromInt byte), j )
				       ) (parseU1_i v i)
			      |  59 => ( Jistore_0, i )
			      |  60 => ( Jistore_1, i )
			      |  61 => ( Jistore_2, i )
			      |  62 => ( Jistore_3, i )
			      |  63 => ( Jlstore_0, i )
			      |  64 => ( Jlstore_1, i )
			      |  65 => ( Jlstore_2, i )
			      |  66 => ( Jlstore_3, i )
			      |  67 => ( Jfstore_0, i )
			      |  68 => ( Jfstore_1, i )
			      |  69 => ( Jfstore_2, i )
			      |  70 => ( Jfstore_3, i )
			      |  71 => ( Jdstore_0, i )
			      |  72 => ( Jdstore_1, i )
			      |  73 => ( Jdstore_2, i )
			      |  74 => ( Jdstore_3, i )
			      |  75 => ( Jastore_0, i )
			      |  76 => ( Jastore_1, i )
			      |  77 => ( Jastore_2, i )
			      |  78 => ( Jastore_3, i )
			      |  79 => ( Jiastore , i )
			      |  80 => ( Jlastore , i )
			      |  81 => ( Jfastore , i )
			      |  82 => ( Jdastore , i )
			      |  83 => ( Jaastore , i )
			      |  84 => ( Jbastore , i )
			      |  85 => ( Jcastore , i )
			      |  86 => ( Jsastore , i )
			      |  87 => ( Jpop     , i )
			      |  88 => ( Jpop2    , i )
			      |  89 => ( Jdup     , i )
			      |  90 => ( Jdup_x1  , i )
			      |  91 => ( Jdup_x2  , i )
			      |  92 => ( Jdup2    , i )
			      |  93 => ( Jdup2_x1 , i )
			      |  94 => ( Jdup2_x2 , i )
			      |  95 => ( Jswap    , i )
			      |  96 => ( Jiadd    , i )
			      |  97 => ( Jladd    , i )
			      |  98 => ( Jfadd    , i )
			      |  99 => ( Jdadd    , i )
			      | 100 => ( Jisub    , i )
			      | 101 => ( Jlsub    , i )
			      | 102 => ( Jfsub    , i )
			      | 103 => ( Jdsub    , i )
			      | 104 => ( Jimul    , i )
			      | 105 => ( Jlmul    , i )
			      | 106 => ( Jfmul    , i )
			      | 107 => ( Jdmul    , i )
			      | 108 => ( Jidiv    , i )
			      | 109 => ( Jldiv    , i )
			      | 110 => ( Jfdiv    , i )
			      | 111 => ( Jddiv    , i )
			      | 112 => ( Jirem    , i )
			      | 113 => ( Jlrem    , i )
			      | 114 => ( Jfrem    , i )
			      | 115 => ( Jdrem    , i )
			      | 116 => ( Jineg    , i )
			      | 117 => ( Jlneg    , i )
			      | 118 => ( Jfneg    , i )
			      | 119 => ( Jdneg    , i )
			      | 120 => ( Jishl    , i )
			      | 121 => ( Jlshl    , i )
			      | 122 => ( Jishr    , i )
			      | 123 => ( Jlshr    , i )
			      | 124 => ( Jiushr   , i )
			      | 125 => ( Jlushr   , i )
			      | 126 => ( Jiand    , i )
			      | 127 => ( Jland    , i )
			      | 128 => ( Jior     , i )
			      | 129 => ( Jlor     , i )
			      | 130 => ( Jixor    , i )
			      | 131 => ( Jlxor    , i )
			      | 132 => let val (byte1,i) = parseU1_i v i 
				           val (byte2,i) = parseU1_i v i
				       in  ( Jiinc { var   = fromInt byte1
				                   , const = byte2 }
				           , i )
				       end
			      | 133 => ( Ji2l     , i )
			      | 134 => ( Ji2f     , i )
			      | 135 => ( Ji2d     , i )
			      | 136 => ( Jl2i     , i )
			      | 137 => ( Jl2f     , i )
			      | 138 => ( Jl2d     , i )
			      | 139 => ( Jf2i     , i )
			      | 140 => ( Jf2l     , i )
			      | 141 => ( Jf2d     , i )
			      | 142 => ( Jd2i     , i )
			      | 143 => ( Jd2l     , i )
			      | 144 => ( Jd2f     , i )
			      | 145 => ( Ji2b     , i )
			      | 146 => ( Ji2c     , i )
			      | 147 => ( Ji2s     , i )
			      | 148 => ( Jlcmp    , i )
			      | 149 => ( Jfcmpl   , i )
			      | 150 => ( Jfcmpg   , i )
			      | 151 => ( Jdcmpl   , i )
			      | 152 => ( Jdcmpg   , i )
			      | 153 => ( fn (bytes2,j) => ( Jifeq      (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)				       
			      | 154 => ( fn (bytes2,j) => ( Jifne      (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)
			      | 155 => ( fn (bytes2,j) => ( Jiflt      (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 156 => ( fn (bytes2,j) => ( Jifge      (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 157 => ( fn (bytes2,j) => ( Jifgt      (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 158 => ( fn (bytes2,j) => ( Jifle      (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 159 => ( fn (bytes2,j) => ( Jif_icmpeq (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 160 => ( fn (bytes2,j) => ( Jif_icmpne (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 161 => ( fn (bytes2,j) => ( Jif_icmplt (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 162 => ( fn (bytes2,j) => ( Jif_icmpge (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 163 => ( fn (bytes2,j) => ( Jif_icmpgt (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 164 => ( fn (bytes2,j) => ( Jif_icmple (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 165 => ( fn (bytes2,j) => ( Jif_acmpeq (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 166 => ( fn (bytes2,j) => ( Jif_acmpne (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)	       
			      | 167 => ( fn (bytes2,j) => ( Jgoto      (Label.fromInt (bytes2)), j )
				       ) (parseU2_offset v i)	       
			      | 168 => ( fn (bytes2,j) => ( Jjsr       (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)		       
			      | 169 => ( fn (byte  ,j) => ( Jret       (fromInt byte)        , j )
				       ) (parseU1_i v i)			       
		              | 170 => let val offset_size   = Int.rem (i,4)
			                   val (offset,j)    = <**> parseU1_i ([],i,offset_size) v
					   val (default,j)   = parseU4_i v j
					   val (low,j)       = parseU4_i v j
					   val (high,j)      = parseU4_i v j
					   val size          = high - low + 1
				           val (targets,j)   = <**> parseU4_i ([],j,size) v
					   fun toLab t       = Label.fromInt (i-1+t)
					   val t_vec         = Vector.fromList (map toLab targets)
				       in  ( Jtableswitch { default = Label.fromInt (i-1+default)
				                          , offset  = Int32.fromInt (i+offset_size+12)
				       	                  , targets = t_vec }
					   , j )
				       end
	 		      | 171 => let val offset_size = 4 - (i div 4)
			                   val (offset,j)  = <**> parseU1_i ([],i,offset_size) v
					   val (default,j) = parseU4_i v j 
					   val (npairs,j)  = parseU4_i v j
					   fun pPair v n   = ( fn (r1,n1) =>
					                       ( fn (r2,n2) =>
							         ( (r1,Label.fromInt (i-1+r2)) , n2 )
							       ) (parseU4_i v n1)
							     ) (parseU4_i32 v n)
				           val (targets,j) = <**> pPair ([],j,npairs) v
				       in  ( Jlookupswitch { default = Label.fromInt (i-1+default)
			                                   , cases = targets  }
					   , j )
				       end
			      | 172 => ( Jireturn , i )
			      | 173 => ( Jlreturn , i )
			      | 174 => ( Jfreturn , i )
			      | 175 => ( Jdreturn , i )
			      | 176 => ( Jareturn , i )
			      | 177 => ( Jreturn  , i )
			      | 178 => let val (idx,i) = parseU2_idx v i
				           val fld_ref = getField_ref pool idx
				       in  ( Jgetstatic     fld_ref , i )
				       end
			      | 179 => let val (idx,i) = parseU2_idx v i
				           val fld_ref = getField_ref pool idx
				       in  ( Jputstatic     fld_ref , i )
				       end
			      | 180 => let val (idx,i) = parseU2_idx v i
				           val fld_ref = getField_ref pool idx
				       in  ( Jgetfield      fld_ref , i )
				       end
			      | 181 => let val (idx,i) = parseU2_idx v i
				           val fld_ref = getField_ref pool idx
				       in  ( Jputfield      fld_ref , i )
				       end
			      | 182 => let val (idx,i) = parseU2_idx v i
				           val mthd_ref = getMethod_ref pool idx
				       in  ( Jinvokevirtual mthd_ref, i )
				       end
			      | 183 => let val (idx,i) = parseU2_idx v i
				           val mthd_ref = getMethod_ref pool idx
				       in  ( Jinvokespecial mthd_ref, i )
				       end
			      | 184 => let val (idx,i) = parseU2_idx v i
				           val mthd_ref = getMethod_ref pool idx
				       in  ( Jinvokestatic  mthd_ref, i )
				       end
			      | 185 => let val (idx,i) = parseU2_idx v i
				                    val mthd_ref = getMethod_ref pool idx
						in  ( Jinvokeinterface mthd_ref, i )
						end
(*			      | 186 =>  not in the JVM-set *)
			      | 187 => let val (idx,i) = parseU2_idx v i
				           val cl_name = Constpool.extract_class pool idx
				       in  ( Jnew (Jvmtype.class { pkgs = []
				                                 , name = cl_name }
						  )
					   , i )
						end
			      | 188 => let val (kind_int,i) = parseU1_i v i
				           val kind = case kind_int of
					                 4 => Jvmtype.Tboolean 
						      |  5 => Jvmtype.Tchar
						      |	 6 => Jvmtype.Tfloat
						      |	 7 => Jvmtype.Tdouble
						      |	 8 => Jvmtype.Tbyte
						      |	 9 => Jvmtype.Tshort
						      |	10 => Jvmtype.Tint
						      |	11 => Jvmtype.Tlong
						      |  n => raise CannotDo "kind IN jvm_instr 188"
				       in ( Jnewarray kind, i )
				       end
			      | 189 => let val (idx,i)   = parseU2_idx v i
				           val cl_name   = Constpool.extract_class pool idx
					   val char_list =  explode cl_name
					   val fst_char  = hd char_list
					   val rest      = tl char_list
				       in  case fst_char of
					     #"[" => ( Janewarray ((ARRAY o #1 o parseType_aux) rest)
						     , i ) 
					   | _    => ( Janewarray ( (CLASS o Jvmtype.class)
					                            { pkgs = []
					                            , name = cl_name }
								  )
						     , i )
				       end
			      | 190 => ( Jarraylength, i )
			      | 191 => ( Jathrow     , i )
			      | 192 => let val (idx,i) = parseU2_idx v i
				           val cl_ref  = parseClass_ref pool idx
				       in  ( Jcheckcast cl_ref , i )
				       end
			      | 193 => let val (idx,i) = parseU2_idx v i
				           val cl_ref  = parseClass_ref pool idx
				       in  ( Jinstanceof cl_ref, i )
				       end
			      | 194 => ( Jmonitorenter, i )
			      | 195 => ( Jmonitorexit , i )
			      | 196 => ( fn (instr,i) => ( Jwide instr, i )
				       ) (parseCode v i)           (* no distinction between wide/regular ops *)
			      | 197 => let val (idx,i)   = parseU2_idx v i
				           val cl_name = Constpool.extract_class pool idx
				           val (dims, i) = parseU1_i v i
				       in  ( Jmultianewarray { elem = parseType cl_name
  		                                             , dim=dims }
					   , i )
				       end
			      | 198 => ( fn (bytes2,j) => ( Jifnull (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)				       
			      | 199 => ( fn (bytes2,j) => ( Jifnonnull (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)				       
			      | 200 => ( fn (bytes4,j) => ( Jgoto_w (Label.fromInt (bytes4)), j )
				       ) (parseU4_i (*should be: parseU4_i offset*) v i)   
			      | 201 => ( fn (bytes2,j) => ( Jjsr_w (Label.fromInt bytes2), j )
				       ) (parseU2_offset v i)				       
			      | n   => raise (UnknownInstructionError n)
			end
       val (bytecode_list, j) = <***> parseCode ([], 0, size) v
   in  bytecode_list
   end
end
























local
   open Bytecode2 Localvar
in 
   fun parseByteCode2 v =
   let val size      = Word8Vector.length v
       val array_ref = ref (Array.array (size,Jempty))
       fun parseCode v i = let val index     = i
                               val (instr,i) = parseU1_i v i
			       fun update triple new_i = ((Array.update triple; #3 triple), new_i )
                           in case instr of
			          0 => update (!array_ref,index,Jnop) i
			      |   1 => update (!array_ref,index,Jaconst_null) i
			      |   2 => update (!array_ref,index,Jiconst_m1) i
			      |   3 => update (!array_ref,index,Jiconst_0) i
			      |   4 => update (!array_ref,index,Jiconst_1) i
			      |   5 => update (!array_ref,index,Jiconst_2) i
			      |   6 => update (!array_ref,index,Jiconst_3) i
			      |   7 => update (!array_ref,index,Jiconst_4) i
			      |   8 => update (!array_ref,index,Jiconst_5) i
			      |   9 => update (!array_ref,index,Jlconst_0) i
			      |  10 => update (!array_ref,index,Jlconst_1) i
			      |  11 => update (!array_ref,index,Jfconst_0) i
			      |  12 => update (!array_ref,index,Jfconst_1) i
			      |  13 => update (!array_ref,index,Jfconst_2) i
			      |  14 => update (!array_ref,index,Jdconst_0) i
			      |  15 => update (!array_ref,index,Jdconst_1) i
			      |  16 => ( fn (byte,j) => ( update (!array_ref,index,Jbipush byte) j )
				       ) (parseU1_i v i)
			      |  17 => ( fn (short,j) => ( update (!array_ref,index,Jsipush short) j )
				       ) (parseU2_i v i)
			      |  18 => let val (cp_idx, i) = (parseU1_idx v i)
                                       in  ( update (!array_ref,index,Jldc cp_idx) i)
				       end
			      |  19 => let val (cp_idx, i) = (parseU2_idx v i)
                                       in  ( update (!array_ref,index,Jldc_w cp_idx) i)
				       end
			      |  20 => let val (cp_idx, i) = (parseU2_idx v i)
                                       in  (update (!array_ref,index,Jldc2_w cp_idx) i)
				       end
			      |  21 => ( fn (byte,j) => ( update (!array_ref,index,Jiload (fromInt byte)) j )
				       ) (parseU1_i v i)
			      |  22 => ( fn (byte,j) => ( update (!array_ref,index,Jlload (fromInt byte)) j )
				       ) (parseU1_i v i)
			      |  23 => ( fn (byte,j) => ( update (!array_ref,index,Jfload (fromInt byte)) j )
				       ) (parseU1_i v i)
			      |  24 => ( fn (byte,j) => ( update (!array_ref,index,Jdload (fromInt byte)) j )
				       ) (parseU1_i v i)
			      |  25 => ( fn (byte,j) => ( update (!array_ref,index,Jaload (fromInt byte)) j )
				       ) (parseU1_i v i)
			      |  26 => ( update (!array_ref,index,Jiload_0) i )
			      |  27 => ( update (!array_ref,index,Jiload_1) i )
			      |  28 => ( update (!array_ref,index,Jiload_2) i )
			      |  29 => ( update (!array_ref,index,Jiload_3) i )
			      |  30 => ( update (!array_ref,index,Jlload_0) i )
			      |  31 => ( update (!array_ref,index,Jlload_1) i )
			      |  32 => ( update (!array_ref,index,Jlload_2) i )
			      |  33 => ( update (!array_ref,index,Jlload_3) i )
			      |  34 => ( update (!array_ref,index,Jfload_0) i )
			      |  35 => ( update (!array_ref,index,Jfload_1) i )
			      |  36 => ( update (!array_ref,index,Jfload_2) i )
			      |  37 => ( update (!array_ref,index,Jfload_3) i )
			      |  38 => ( update (!array_ref,index,Jdload_0) i )
			      |  39 => ( update (!array_ref,index,Jdload_1) i )
			      |  40 => ( update (!array_ref,index,Jdload_2) i )
			      |  41 => ( update (!array_ref,index,Jdload_3) i )
			      |  42 => ( update (!array_ref,index,Jaload_0) i ) 
			      |  43 => ( update (!array_ref,index,Jaload_1) i )
			      |  44 => ( update (!array_ref,index,Jaload_2) i )
			      |  45 => ( update (!array_ref,index,Jaload_3) i )
			      |  46 => ( update (!array_ref,index,Jiaload) i )
			      |  47 => ( update (!array_ref,index,Jlaload) i )
			      |  48 => ( update (!array_ref,index,Jfaload) i )
			      |  49 => ( update (!array_ref,index,Jdaload) i )
			      |  50 => ( update (!array_ref,index,Jaaload) i )
			      |  51 => ( update (!array_ref,index,Jbaload) i )
			      |  52 => ( update (!array_ref,index,Jcaload) i )
			      |  53 => ( update (!array_ref,index,Jsaload) i )
			      |  54 => ( fn (byte,j) => ( update (!array_ref,index,Jistore (fromInt byte)) j )
				       ) (parseU1_i v i)
			      |  55 => ( fn (byte,j) => ( update (!array_ref,index,Jlstore (fromInt byte)) j )
				       ) (parseU1_i v i)
			      |  56 => ( fn (byte,j) => ( update (!array_ref,index,Jfstore (fromInt byte)) j )
				       ) (parseU1_i v i)
			      |  57 => ( fn (byte,j) => ( update (!array_ref,index,Jdstore (fromInt byte)) j )
				       ) (parseU1_i v i)
			      |  58 => ( fn (byte,j) => ( update (!array_ref,index,Jastore (fromInt byte)) j )
				       ) (parseU1_i v i)
			      |  59 => ( update (!array_ref,index,Jistore_0) i )
			      |  60 => ( update (!array_ref,index,Jistore_1) i )
			      |  61 => ( update (!array_ref,index,Jistore_2) i )
			      |  62 => ( update (!array_ref,index,Jistore_3) i )
			      |  63 => ( update (!array_ref,index,Jlstore_0) i )
			      |  64 => ( update (!array_ref,index,Jlstore_1) i )
			      |  65 => ( update (!array_ref,index,Jlstore_2) i )
			      |  66 => ( update (!array_ref,index,Jlstore_3) i )
			      |  67 => ( update (!array_ref,index,Jfstore_0) i )
			      |  68 => ( update (!array_ref,index,Jfstore_1) i )
			      |  69 => ( update (!array_ref,index,Jfstore_2) i )
			      |  70 => ( update (!array_ref,index,Jfstore_3) i )
			      |  71 => ( update (!array_ref,index,Jdstore_0) i )
			      |  72 => ( update (!array_ref,index,Jdstore_1) i )
			      |  73 => ( update (!array_ref,index,Jdstore_2) i )
			      |  74 => ( update (!array_ref,index,Jdstore_3) i )
			      |  75 => ( update (!array_ref,index,Jastore_0) i )
			      |  76 => ( update (!array_ref,index,Jastore_1) i )
			      |  77 => ( update (!array_ref,index,Jastore_2) i )
			      |  78 => ( update (!array_ref,index,Jastore_3) i )
			      |  79 => ( update (!array_ref,index,Jiastore) i )
			      |  80 => ( update (!array_ref,index,Jlastore) i )
			      |  81 => ( update (!array_ref,index,Jfastore) i )
			      |  82 => ( update (!array_ref,index,Jdastore) i )
			      |  83 => ( update (!array_ref,index,Jaastore) i )
			      |  84 => ( update (!array_ref,index,Jbastore) i )
			      |  85 => ( update (!array_ref,index,Jcastore) i )
			      |  86 => ( update (!array_ref,index,Jsastore) i )
			      |  87 => ( update (!array_ref,index,Jpop) i )
			      |  88 => ( update (!array_ref,index,Jpop2) i )
			      |  89 => ( update (!array_ref,index,Jdup) i )
			      |  90 => ( update (!array_ref,index,Jdup_x1) i )
			      |  91 => ( update (!array_ref,index,Jdup_x2) i )
			      |  92 => ( update (!array_ref,index,Jdup2) i )
			      |  93 => ( update (!array_ref,index,Jdup2_x1) i )
			      |  94 => ( update (!array_ref,index,Jdup2_x2) i )
			      |  95 => ( update (!array_ref,index,Jswap) i )
			      |  96 => ( update (!array_ref,index,Jiadd) i )
			      |  97 => ( update (!array_ref,index,Jladd) i )
			      |  98 => ( update (!array_ref,index,Jfadd) i )
			      |  99 => ( update (!array_ref,index,Jdadd) i )
			      | 100 => ( update (!array_ref,index,Jisub) i )
			      | 101 => ( update (!array_ref,index,Jlsub) i )
			      | 102 => ( update (!array_ref,index,Jfsub) i )
			      | 103 => ( update (!array_ref,index,Jdsub) i )
			      | 104 => ( update (!array_ref,index,Jimul) i )
			      | 105 => ( update (!array_ref,index,Jlmul) i )
			      | 106 => ( update (!array_ref,index,Jfmul) i )
			      | 107 => ( update (!array_ref,index,Jdmul) i )
			      | 108 => ( update (!array_ref,index,Jidiv) i )
			      | 109 => ( update (!array_ref,index,Jldiv) i )
			      | 110 => ( update (!array_ref,index,Jfdiv) i )
			      | 111 => ( update (!array_ref,index,Jddiv) i )
			      | 112 => ( update (!array_ref,index,Jirem) i )
			      | 113 => ( update (!array_ref,index,Jlrem) i )
			      | 114 => ( update (!array_ref,index,Jfrem) i )
			      | 115 => ( update (!array_ref,index,Jdrem) i )
			      | 116 => ( update (!array_ref,index,Jineg) i )
			      | 117 => ( update (!array_ref,index,Jlneg) i )
			      | 118 => ( update (!array_ref,index,Jfneg) i )
			      | 119 => ( update (!array_ref,index,Jdneg) i )
			      | 120 => ( update (!array_ref,index,Jishl) i )
			      | 121 => ( update (!array_ref,index,Jlshl) i )
			      | 122 => ( update (!array_ref,index,Jishr) i )
			      | 123 => ( update (!array_ref,index,Jlshr) i )
			      | 124 => ( update (!array_ref,index,Jiushr) i )
			      | 125 => ( update (!array_ref,index,Jlushr) i )
			      | 126 => ( update (!array_ref,index,Jiand) i )
			      | 127 => ( update (!array_ref,index,Jland) i )
			      | 128 => ( update (!array_ref,index,Jior) i )
			      | 129 => ( update (!array_ref,index,Jlor) i )
			      | 130 => ( update (!array_ref,index,Jixor) i )
			      | 131 => ( update (!array_ref,index,Jlxor) i )
			      | 132 => let val (byte1,i) = parseU1_i v i 
				           val (byte2,i) = parseU1_i v i
				       in  ( update ( !array_ref,index,Jiinc { var   = fromInt byte1
				                      			     , const = byte2 } )
					              i )
				       end
			      | 133 => ( update (!array_ref,index,Ji2l) i )
			      | 134 => ( update (!array_ref,index,Ji2f) i )
			      | 135 => ( update (!array_ref,index,Ji2d) i )
			      | 136 => ( update (!array_ref,index,Jl2i) i )
			      | 137 => ( update (!array_ref,index,Jl2f) i )
			      | 138 => ( update (!array_ref,index,Jl2d) i )
			      | 139 => ( update (!array_ref,index,Jf2i) i )
			      | 140 => ( update (!array_ref,index,Jf2l) i )
			      | 141 => ( update (!array_ref,index,Jf2d) i )
			      | 142 => ( update (!array_ref,index,Jd2i) i )
			      | 143 => ( update (!array_ref,index,Jd2l) i )
			      | 144 => ( update (!array_ref,index,Jd2f) i )
			      | 145 => ( update (!array_ref,index,Ji2b) i )
			      | 146 => ( update (!array_ref,index,Ji2c) i )
			      | 147 => ( update (!array_ref,index,Ji2s) i )
			      | 148 => ( update (!array_ref,index,Jlcmp) i )
			      | 149 => ( update (!array_ref,index,Jfcmpl) i )
			      | 150 => ( update (!array_ref,index,Jfcmpg) i )
			      | 151 => ( update (!array_ref,index,Jdcmpl) i )
			      | 152 => ( update (!array_ref,index,Jdcmpg) i )
			      | 153 => ( fn (bytes2,j) => ( update (!array_ref,index,Jifeq (OFF bytes2)) j )
				       ) (parseU2_offset v i)				       
			      | 154 => ( fn (bytes2,j) => ( update (!array_ref,index,Jifne (OFF bytes2)) j )
				       ) (parseU2_offset v i)
			      | 155 => ( fn (bytes2,j) => ( update (!array_ref,index,Jiflt (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 156 => ( fn (bytes2,j) => ( update (!array_ref,index,Jifge (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 157 => ( fn (bytes2,j) => ( update (!array_ref,index,Jifgt (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 158 => ( fn (bytes2,j) => ( update (!array_ref,index,Jifle (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 159 => ( fn (bytes2,j) => ( update (!array_ref,index,Jif_icmpeq (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 160 => ( fn (bytes2,j) => ( update (!array_ref,index,Jif_icmpne (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 161 => ( fn (bytes2,j) => ( update (!array_ref,index,Jif_icmplt (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 162 => ( fn (bytes2,j) => ( update (!array_ref,index,Jif_icmpge (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 163 => ( fn (bytes2,j) => ( update (!array_ref,index,Jif_icmpgt (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 164 => ( fn (bytes2,j) => ( update (!array_ref,index,Jif_icmple (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 165 => ( fn (bytes2,j) => ( update (!array_ref,index,Jif_acmpeq (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 166 => ( fn (bytes2,j) => ( update (!array_ref,index,Jif_acmpne (OFF bytes2)) j )
				       ) (parseU2_offset v i)	       
			      | 167 => ( fn (bytes2,j) => ( update (!array_ref,index,Jgoto (OFF bytes2)) j )
				       ) (parseU2_offset v i)
			      | 168 => ( fn (bytes2,j) => ( update (!array_ref,index,Jjsr (OFF bytes2)) j )
				       ) (parseU2_offset v i)		       
			      | 169 => ( fn (byte  ,j) => ( update (!array_ref,index,Jret (fromInt byte)) j )
				       ) (parseU1_i v i)			       
		              | 170 => let val offset_size   = Int.rem (i,4)
			                   val (offset,j)    = <**> parseU1_i ([],i,offset_size) v
					   val (default,j)   = parseU4_i v j
					   val (low,j)       = parseU4_i v j
					   val (high,j)      = parseU4_i v j
					   val size          = high - low + 1
				           val (targets,j)   = <**> parseU4_i ([],j,size) v
					   fun toLab t       = OFF t
					   val t_vec         = Vector.fromList (map toLab targets)
				       in  ( update (!array_ref,index,
				             Jtableswitch { default = OFF default
				                          , low     = low
							  , high    = high
							  , offsets = t_vec (*label Vector.vector*) }
					   ) j )
				       end
	 		      | 171 => let val offset_size = 4 - (i div 4)
			                   val (offset,j)  = <**> parseU1_i ([],i,offset_size) v
					   val (default,j) = parseU4_i v j 
					   val (npairs,j)  = parseU4_i v j
					   fun pPair v n   = ( fn (r1,n1) =>
					                       ( fn (r2,n2) =>
							         ( (r1,OFF r2) , n2 )
							       ) (parseU4_i v n1)
							     ) (parseU4_i32 v n)
				           val (targets,j) = <**> pPair ([],j,npairs) v
				       in  ( update (!array_ref,index,
				             Jlookupswitch { default = OFF default
			                                   , cases = targets }
					   ) j )
				       end
			      | 172 => ( update (!array_ref,index,Jireturn) i )
			      | 173 => ( update (!array_ref,index,Jlreturn) i )
			      | 174 => ( update (!array_ref,index,Jfreturn) i )
			      | 175 => ( update (!array_ref,index,Jdreturn) i )
			      | 176 => ( update (!array_ref,index,Jareturn) i )
			      | 177 => ( update (!array_ref,index,Jreturn) i )
			      | 178 => let val (cp_idx,i) = parseU2_idx v i
				       in  ( update (!array_ref,index,Jgetstatic cp_idx) i )
				       end
			      | 179 => let val (cp_idx,i) = parseU2_idx v i
				       in  ( update (!array_ref,index,Jputstatic cp_idx) i )
				       end
			      | 180 => let val (cp_idx,i) = parseU2_idx v i
				       in  ( update (!array_ref,index,Jgetfield cp_idx) i )
				       end
			      | 181 => let val (cp_idx,i) = parseU2_idx v i
				       in  ( update (!array_ref,index,Jputfield cp_idx) i )
				       end
			      | 182 => let val (cp_idx,i) = parseU2_idx v i
				       in  ( update (!array_ref,index,Jinvokevirtual cp_idx) i )
				       end
			      | 183 => let val (cp_idx,i) = parseU2_idx v i
				       in  ( update (!array_ref,index,Jinvokespecial cp_idx) i )
				       end
			      | 184 => let val (cp_idx,i) = parseU2_idx v i
				       in  ( update (!array_ref,index,Jinvokestatic  cp_idx) i )
				       end
			      | 185 => let val (cp_idx,i) = parseU2_idx v i
				       in  ( update (!array_ref,index,Jinvokeinterface cp_idx) i )
				       end
(*			      | 186 =>  not in the JVM-set *)
			      | 187 => let val (cp_idx,i) = parseU2_idx v i
				       in  ( update (!array_ref,index,Jnew cp_idx) i )
				       end
			      | 188 => let val (kind_int,i) = parseU1_i v i
				           val kind = case kind_int of
					                 4 => Abool 
						      |  5 => Achar
						      |	 6 => Afloat
						      |	 7 => Adouble
						      |	 8 => Abyte
						      |	 9 => Ashort
						      |	10 => Aint
						      |	11 => Along
						      |  n => raise CannotDo "kind IN jvm_instr 188"
				       in ( update (!array_ref,index,Jnewarray kind) i )
				       end
			      | 189 => let val (cp_idx,i)   = parseU2_idx v i
				       in  ( update (!array_ref,index,Janewarray cp_idx) i ) 
				       end
			      | 190 => ( update (!array_ref,index,Jarraylength) i )
			      | 191 => ( update (!array_ref,index,Jathrow) i )
			      | 192 => let val (cp_idx,i) = parseU2_idx v i
				       in  ( update (!array_ref,index,Jcheckcast cp_idx) i )
				       end
			      | 193 => let val (cp_idx,i) = parseU2_idx v i
				       in  ( update (!array_ref,index,Jinstanceof cp_idx) i )
				       end
			      | 194 => ( update (!array_ref,index,Jmonitorenter) i )
			      | 195 => ( update (!array_ref,index,Jmonitorexit) i )
			      | 196 => ( fn (instr,i) => ( update (!array_ref,index,Jwide instr) i )
				       ) (parseCode v i)           (* no distinction between wide/regular ops *)
			      | 197 => let val (cp_idx,i)   = parseU2_idx v i
				           val (dims, i) = parseU1_i v i
				       in  ( update (!array_ref,index,
				             Jmultianewarray { elem = cp_idx
  		                                             , dim  = dims }
					   ) i )
				       end
			      | 198 => ( fn (bytes2,j) => ( update (!array_ref,index,Jifnull (OFF bytes2)) j )
				       ) (parseU2_offset v i)				       
			      | 199 => ( fn (bytes2,j) => ( update (!array_ref,index,Jifnonnull (OFF bytes2)) j )
				       ) (parseU2_offset v i)				       
			      | 200 => ( fn (bytes4,j) => ( update (!array_ref,index,Jgoto_w (OFF bytes4)) j )
				       ) (parseU4_i v i) (*TODO: parseU4_i offset*)
			      | 201 => ( fn (bytes2,j) => ( update (!array_ref,index,Jjsr_w (OFF bytes2)) j )
				       ) (parseU2_offset v i)				       
			      | n   => raise (UnknownInstructionError n)
			end
       val (bytecode_list, j) = <***> parseCode ([], 0, size) v
   in  !array_ref
   end
end







end
