(*                                                                      *)
(* MLJVM - a Java Virtual Machine implemented in Standard ML            *)
(*                                                                      *)
(* http://homepages.inf.ed.ac.uk/stg/research/projects/MLJVM/           *)
(*                                                                      *)
(* Copyright (C) 2003 Stephen Gilmore and Laura Korte, The              *)
(* University of Edinburgh.                                             *)
(*                                                                      *)
(* This program is free software; you can redistribute it and/or        *)
(* modify it under the terms of the GNU General Public License          *)
(* as published by the Free Software Foundation; either version 2       *)
(* of the License, or (at your option) any later version.               *)
(*                                                                      *)
(* This program is distributed in the hope that it will be useful,      *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of       *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *)
(* GNU General Public License for more details.                         *)
(*                                                                      *)
(* You should have received a copy of the GNU General Public License    *)
(* along with this program; if not, write to the Free Software          *)
(* Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *)
(* MA 02111-1307, USA.                                                  *)
(*                                                                      *)


structure RuntimeClass =
struct

exception Error of string;
val file = "RuntimeClass.sml"
fun error s = Error (s ^ " in " ^ file)

local
   open Label Bytecode2
in
   
   type local_var_info =
        { start  : label
        , length : label
        , name   : RuntimePool.index
        , desc   : RuntimePool.index
        , index  : Localvar.index }
 
   type line_number_info =
        { start : label
        , line  : int }

   type exn_hdl =
        { start : label
        , stop  : label
        , entry : label
        , catch : RuntimePool.index option }

   type inner_class_info =
        { inner_info  : RuntimePool.index option
        , outer_info  : RuntimePool.index option
        , inner_name  : RuntimePool.index option
        , inner_flags : Modifiers.access_flag list } 

   datatype value = Vempty (*placeholder*)
                  | Vint of Int32.int
                  | Vlong of Int64.int
		  | Vfloat of Real32.real
		  | Vdouble of Real64.real
		  | Vobj of int (*object_pointer*)
   
   datatype attribute =
       EMPTY    (*placeholder*)
     | SRCFILE  of RuntimePool.index
     | CONSTVAL of RuntimePool.index
     | CODE     of { stack  : int
 	           , locals : int
	           , code   : jvm_instr Array.array
	           , hdls   : exn_hdl list
	           , attrs  : attribute Array.array }
     | EXNS     of RuntimePool.index list
     | LINENUM  of line_number_info list
     | LOCALVAR of local_var_info list
     | INNER    of inner_class_info list
     | ATTR     of { attr : RuntimePool.index
  	           , info : Word8Vector.vector}
     | VALUE    of value

   (*attribute-indices into the attribute-arrays*)
   val srcfile  = 0
   val constval = 1
   val code     = 2
   val exns     = 3
   val linenum  = 4
   val localvar = 5
   val inner    = 6
   val attr     = 7
   val value    = 8

   type member =
        { flags : Modifiers.access_flag list
        , name  : RuntimePool.index
        , desc  : RuntimePool.index
        , attrs : attribute Array.array }

   type class =
        { class_ref : int (*Object.pointer to object representing class*)
	, pool      : RuntimePool.pool
        , flags     : Modifiers.access_flag list
        , this      : RuntimePool.index
        , super     : RuntimePool.index option
        , ifcs      : RuntimePool.index list
        , fields    : { array : (member option) Array.array
	              , hash  : (string, int) Polyhash.hash_table }
        , methods   : { array : (member option) Array.array
	              , hash  : (string, int) Polyhash.hash_table }
        , attrs     : attribute Array.array }
   
   
   type method_area = { array : (class option) Array.array
                      , hash  : (string,int) Polyhash.hash_table }

   (*************************************************)   
   (************* start of native classes ***********)   
   (*************************************************)   

   fun empty_attr_array () = Array.array (9,EMPTY)
   fun empty_instr_array n = Array.array (n,Jempty)
   fun empty_member_array n = Array.array (n,NONE)
   val findExn = error "Error while trying to do Polyhash.find"
   fun empty_member_hash n = Polyhash.mkPolyTable (n,findExn)
   fun empty_tuple n = { array = empty_member_array n
                       , hash  = empty_member_hash n }
   fun updateTuple ({array, hash},n,item) s =
       let val _ =  Array.update (array,n,item)
           val _ =  Polyhash.insert hash (s,n)
       in { array = array
          , hash  = hash }
       end
   
   local
      val instr_array = empty_instr_array 1
      val _ = Array.update ( instr_array, 0, Jreturn )
      val attr_array = empty_attr_array ()
      val _ = Array.update ( attr_array
                           , code 
			   , CODE { stack  = 3
 	                          , locals = 3
	                          , code   = instr_array
	                          , hdls   = []
	                          , attrs  = empty_attr_array () } )
   in 
      val object_init =
          { flags = []
          , name  = RuntimePool.makeIndex 3
          , desc  = RuntimePool.makeIndex 4
          , attrs = attr_array } : member
   end

   local
      open RuntimePool
      val method_tuple = empty_tuple 1
      val method_tuple = updateTuple (method_tuple,0,SOME object_init) "<init>()V"
   in val object_class = { class_ref = 0
                         , pool      = Array.fromList [ CP (Constpool.CPunused)
		                                      , CP (Constpool.CPclass (Constpool.makeIndex 2))
						      , CP (Constpool.CPutf8 "java/lang/Object")
						      , CP (Constpool.CPutf8 "<init>")
						      , CP (Constpool.CPutf8 "()V") ]
		         , flags     = []
		         , this      = makeIndex 1
		         , super     = NONE
		         , ifcs      = []
		         , fields    = empty_tuple 0
		         , methods   = method_tuple
		         , attrs     = empty_attr_array () } : class
   end

   local
      val attr_array = empty_attr_array ()
      val _ = Array.update ( attr_array
                           , value
			   , VALUE Vempty )
   in 
      val printstream_out =
          { flags = []
          , name  = RuntimePool.makeIndex 3
          , desc  = RuntimePool.makeIndex 4
          , attrs = attr_array } : member
   end

   local
      open RuntimePool
      val field_tuple = empty_tuple 1
      val field_tuple = updateTuple (field_tuple,0,SOME printstream_out) "outLjava/io/PrintStream;"
   in val printstream_class = { class_ref = 0
                              , pool      = Array.fromList [ CP (Constpool.CPunused)
		                                           , CP (Constpool.CPclass (Constpool.makeIndex 2))
						           , CP (Constpool.CPutf8 "java/lang/System")
						           , CP (Constpool.CPutf8 "out")
						           , CP (Constpool.CPutf8 "Ljava/io/PrintStream;") ]
		              , flags     = []
		              , this      = makeIndex 1
		              , super     = NONE
		              , ifcs      = []
		              , fields    = field_tuple
		              , methods   = empty_tuple 0
		              , attrs     = empty_attr_array () } : class
   end
   
   local
      val instr_array = empty_instr_array 2
      val _ = Array.update ( instr_array, 0, Nprintln )
      val _ = Array.update ( instr_array, 1, Jreturn )
      val attr_array = empty_attr_array ()
      val _ = Array.update ( attr_array
                           , code
			   , CODE { stack  = 3
 	                          , locals = 3
	                          , code   = instr_array
	                          , hdls   = []
	                          , attrs  = empty_attr_array () } )
   in 
      val println =
          { flags = []
          , name  = RuntimePool.makeIndex 3
          , desc  = RuntimePool.makeIndex 4
          , attrs = attr_array } : member
   end
   local
      val instr_array = empty_instr_array 2
      val _ = Array.update ( instr_array, 0, Nprint )
      val _ = Array.update ( instr_array, 1, Jreturn )
      val attr_array = empty_attr_array ()
      val _ = Array.update ( attr_array
                           , code
			   , CODE { stack  = 3
 	                          , locals = 3
	                          , code   = instr_array
	                          , hdls   = []
	                          , attrs  = empty_attr_array () } )
   in 
      val print_f =
          { flags = []
          , name  = RuntimePool.makeIndex 5
          , desc  = RuntimePool.makeIndex 4
          , attrs = attr_array } : member
   end
   local
      val instr_array = empty_instr_array 7
      val _ = Array.update ( instr_array, 0, Jiload_1 ) 
      val _ = Array.update ( instr_array, 1, Jinvokestatic (Constpool.makeIndex 7)) 
      val _ = Array.update ( instr_array, 4, Jastore_1 )
      val _ = Array.update ( instr_array, 5, Nprintln )
      val _ = Array.update ( instr_array, 6, Jreturn )
      val attr_array = empty_attr_array ()
      val _ = Array.update ( attr_array
                           , code
			   , CODE { stack  = 3
 	                          , locals = 3
	                          , code   = instr_array
	                          , hdls   = []
	                          , attrs  = empty_attr_array () } )
   in 
      val println_int =
          { flags = []
          , name  = RuntimePool.makeIndex 3
          , desc  = RuntimePool.makeIndex 6
          , attrs = attr_array } : member
   end
   local
      val instr_array = empty_instr_array 7
      val _ = Array.update ( instr_array, 0, Jiload_1 ) 
      val _ = Array.update ( instr_array, 1, Jinvokestatic (Constpool.makeIndex 7)) 
      val _ = Array.update ( instr_array, 4, Jastore_1 )
      val _ = Array.update ( instr_array, 5, Nprint )
      val _ = Array.update ( instr_array, 6, Jreturn )
      val attr_array = empty_attr_array ()
      val _ = Array.update ( attr_array
                           , code
			   , CODE { stack  = 3
 	                          , locals = 3
	                          , code   = instr_array
	                          , hdls   = []
	                          , attrs  = empty_attr_array () } )
   in 
      val print_int =
          { flags = []
          , name  = RuntimePool.makeIndex 5
          , desc  = RuntimePool.makeIndex 6
          , attrs = attr_array } : member
   end

   local
      open RuntimePool
      val method_tuple = empty_tuple 4
      val method_tuple = updateTuple (method_tuple,0,SOME println) "println(Ljava/lang/String;)V"
      val method_tuple = updateTuple (method_tuple,1,SOME print_f) "print(Ljava/lang/String;)V"
      val method_tuple = updateTuple (method_tuple,2,SOME println_int) "println(I)V"
      val method_tuple = updateTuple (method_tuple,3,SOME print_int) "print(I)V"
   in val print_class  = { class_ref = 0
                         , pool      = Array.fromList [ CP (Constpool.CPunused)
		                           (* 01 *)   , CP (Constpool.CPclass (Constpool.makeIndex 2))
					   (* 02 *)   , CP (Constpool.CPutf8 "java/io/PrintStream")
					   (* 03 *)   , CP (Constpool.CPutf8 "println")
					   (* 04 *)   , CP (Constpool.CPutf8 "(Ljava/lang/String;)V")
					   (* 05 *)   , CP (Constpool.CPutf8 "print")
					   (* 06 *)   , CP (Constpool.CPutf8 "(I)V")
					   (* 07 *)   , CP (Constpool.CPmethodref { class = Constpool.makeIndex 8
						                                  , nameType = Constpool.makeIndex 10 } )
					   (* 08 *)   , CP (Constpool.CPclass (Constpool.makeIndex 9))
					   (* 09 *)   , CP (Constpool.CPutf8 "java/lang/Integer")
					   (* 10 *)   , CP (Constpool.CPnametype { name = Constpool.makeIndex 11
						                                 , desc = Constpool.makeIndex 12} )
					   (* 11 *)   , CP (Constpool.CPutf8 "toString")
					   (* 12 *)   , CP (Constpool.CPutf8 "(I)Ljava/lang/String;") ]
		         , flags     = []
		         , this      = makeIndex 1
		         , super     = NONE
		         , ifcs      = []
		         , fields    = empty_tuple 0
		         , methods   = method_tuple
		         , attrs     = empty_attr_array () } : class
   end

   
   
   
   local
      val instr_array = empty_instr_array 2
      val _ = Array.update ( instr_array, 0, Nint2string )
      val _ = Array.update ( instr_array, 1, Jareturn )
      val attr_array = empty_attr_array ()
      val _ = Array.update ( attr_array
                           , code
			   , CODE { stack  = 3
 	                          , locals = 3
	                          , code   = instr_array
	                          , hdls   = []
	                          , attrs  = empty_attr_array () } )
   in 
      val integer_toString =
          { flags = []
          , name  = RuntimePool.makeIndex 3
          , desc  = RuntimePool.makeIndex 4
          , attrs = attr_array } : member
   end

   local
      val instr_array = empty_instr_array 2
      val _ = Array.update ( instr_array, 0, NparseInt )
      val _ = Array.update ( instr_array, 1, Jireturn )
      val attr_array = empty_attr_array ()
      val _ = Array.update ( attr_array
                           , code
			   , CODE { stack  = 3
 	                          , locals = 3
	                          , code   = instr_array
	                          , hdls   = []
	                          , attrs  = empty_attr_array () } )
   in 
      val parseInt =
          { flags = []
          , name  = RuntimePool.makeIndex 5
          , desc  = RuntimePool.makeIndex 6
          , attrs = attr_array } : member
   end


   local
      open RuntimePool
      val method_tuple = empty_tuple 2
      val method_tuple = updateTuple (method_tuple,0,SOME integer_toString) "toString(I)Ljava/lang/String;"
      val method_tuple = updateTuple (method_tuple,1,SOME parseInt) "parseInt(Ljava/lang/String;)I"
   in val integer_class  = { class_ref = 0
                           , pool      = Array.fromList [ CP (Constpool.CPunused)
		                                        , CP (Constpool.CPclass (Constpool.makeIndex 2))
			  			        , CP (Constpool.CPutf8 "java/lang/Integer")
						        , CP (Constpool.CPutf8 "toString")
						        , CP (Constpool.CPutf8 "(I)Ljava/lang/String;")
						        , CP (Constpool.CPutf8 "parseInt")
						        , CP (Constpool.CPutf8 "(Ljava/lang/String;)I") ]
		           , flags     = []
		           , this      = makeIndex 1
		           , super     = NONE
		           , ifcs      = []
		           , fields    = empty_tuple 0
		           , methods   = method_tuple
		           , attrs     = empty_attr_array () } : class
   end


   
   (*************************************************)   
   (************ end of native classes **************)   
   (*************************************************)   
   
      
   datatype findex = FIDX of int
   datatype mindex = MIDX of int
   datatype index  = IDX  of int (*class_index*)
   fun indexValue (IDX i) = i
   
   datatype label  = LBL  of string

   fun findMethodByIndex (m:method_area) (IDX i,MIDX mi) =
       (valOf o Array.sub) ( (#array o #methods o valOf o Array.sub) (#array m,i)
                           , mi ) 
   fun updateFieldByIndex (m:method_area) (IDX i,FIDX fi) v =
       let val array = (#attrs o valOf o Array.sub) ( (#array o #fields o valOf o Array.sub) (#array m,i)
                           , fi )
       in Array.update (array,value,v)
       end

   fun safeValOf NONE = raise (error "safeValOf")
     | safeValOf (SOME x) = x
   fun getFieldValueByIndex (m:method_area) (IDX i,FIDX fi) =
       let val array = (#attrs o safeValOf o Array.sub) ( (#array o #fields o valOf o Array.sub) (#array m,i)
                           , fi )
	   val value = case (Array.sub (array,value)) of
	                    VALUE v => v
			  | _       => raise (error "Match-error while doing getFieldValueByIndex")
       in  value
       end

   fun string2mindex (m:method_area) (IDX i,s) =
       MIDX (Polyhash.find ( ( #hash o #methods o valOf o Array.sub)
                             (#array m,i) ) s) 
   fun string2findex (m:method_area) (IDX i,s) =
       FIDX (Polyhash.find ( ( #hash o #fields o valOf o Array.sub)
                             (#array m,i) ) s) 
   
   fun getCode (CODE code_rec) = #code code_rec
     | getCode _               = raise (error "Expected CODE")
     
   fun findMethodCodeByIndex (m:method_area) (IDX i,MIDX mi) =
       let val m_array = (#array o #methods o valOf o Array.sub) (#array m,i)
           val a_array = (#attrs o valOf o Array.sub) (m_array,mi)
       in  (getCode o Array.sub) (a_array, code)
       end

   fun findCPByLabel (m:method_area) (LBL s) =
       (#pool o valOf o Array.sub) (#array m, Polyhash.find (#hash m) s)
   fun findCPByIndex (m:method_area) (IDX i) =
       (#pool o valOf o Array.sub) (#array m,i)
   fun updateCPByIndex ({array,hash}:method_area) (IDX i) new_pool =
       let val { class_ref
               , pool
	       , flags
	       , this
	       , super
	       , ifcs
	       , fields
	       , methods
	       , attrs } = (valOf o Array.sub) (array,i)
	   val new_class = { class_ref = class_ref
	                   , pool      = new_pool
	       		   , flags     = flags
	       		   , this      = this
	      		   , super     = super
	       		   , ifcs      = ifcs
	       		   , fields    = fields
	       		   , methods   = methods
	       		   , attrs     = attrs }
           val _ = Array.update (array,i,SOME new_class)
       in  {array = array, hash = hash}
       end
   
   fun findClassByLabel (m:method_area) (LBL s) = (valOf o Array.sub) (#array m, Polyhash.find (#hash m) s)
   fun findClassByIndex (m:method_area) (IDX i) = (valOf o Array.sub) (#array m,i)

   fun label2index (m:method_area) (LBL s) = IDX (Polyhash.find (#hash m) s)
   fun label2index_try (m:method_area) (LBL s) = IDX (Polyhash.find (#hash m) s handle _ => 0)

end

end;

