(*                                                                      *)
(* MLJVM - a Java Virtual Machine implemented in Standard ML            *)
(*                                                                      *)
(* http://homepages.inf.ed.ac.uk/stg/research/projects/MLJVM/           *)
(*                                                                      *)
(* Copyright (C) 2003 Stephen Gilmore and Laura Korte, The              *)
(* University of Edinburgh.                                             *)
(*                                                                      *)
(* This program is free software; you can redistribute it and/or        *)
(* modify it under the terms of the GNU General Public License          *)
(* as published by the Free Software Foundation; either version 2       *)
(* of the License, or (at your option) any later version.               *)
(*                                                                      *)
(* This program is distributed in the hope that it will be useful,      *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of       *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *)
(* GNU General Public License for more details.                         *)
(*                                                                      *)
(* You should have received a copy of the GNU General Public License    *)
(* along with this program; if not, write to the Free Software          *)
(* Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *)
(* MA 02111-1307, USA.                                                  *)
(*                                                                      *)

structure Transform :> Transform =
struct

open MyClassFile Constpool

exception Error of string;
exception UnsupportedModifierError of string;
exception UnknownConstpoolConstantValueError;
val findExn = Error "PolyHash-find in Transform.sml failed"

(**********************)
(***** DEBUG INFO *****)
(**********************)

val debug_string = ref "" : string ref
fun do_debug 0 ss = ()
  | do_debug 1 ss = ()
  | do_debug n ss =  ((print ss);(TextIO.flushOut TextIO.stdOut))


(*********************************************************************************************)
(*************************************** auxilary functions **********************************)
(*********************************************************************************************)

fun optional_function f NONE     = NONE
  | optional_function f (SOME s) = SOME (f s)

fun w16_to_label w = Label.fromInt (Word16.toInt w)

(*********************************************************************************************)
(*********************************** file2decl functions *************************************)
(*********************************************************************************************)

(* file2decl flags of: classes, fields, methods *)
fun f2d_modifiers access_flags kind =
let val mod_function = case kind of
                         "class"  => Modifiers.cmod2flag
                       | "method" => Modifiers.mmod2flag
		       | "field"  => Modifiers.fmod2flag
		       | s        => raise UnsupportedModifierError s
    val modifier_list = Modifiers.priority_list
    fun try_mod n (x::xs) = if n-x < 0
                            then ([],n,xs)
			    else ([mod_function x], n-x, xs)
      | try_mod n []      = raise Error "Cannot do f2d_modifiers"
    fun compute_flags (mods_s,flags,l) = if flags > 0
                                         then ( compute_flags
					        o
						( fn ( new_mod          , new_flags, mod_list) =>
					             ( new_mod @ mods_s, new_flags, mod_list)
						)
					      ) (try_mod flags l)
		                         else (mods_s,0)
    val (result,n)  = compute_flags ( [], access_flags, modifier_list )
in  result
end

(*Constpool.index -> Bytecode.jvm_const*)
local open Bytecode
in fun idx2jvm_const pool idx =
   let val cp_item = Constpool.lookup pool idx
   in  case cp_item of
         CPlong   l    => Clong l
       | CPfloat  f    => (Cfloat o valOf o Real32.fromBytes) f
       | CPdouble d    => (Cdouble o valOf o Real64.fromBytes) d
       | CPint    i    => Cint i
       | CPstring idx2 => Cstring (extract_name pool idx2)
       | _             => raise UnknownConstpoolConstantValueError
   end
end

fun idx2jclass pool idx = Jvmtype.class { name = Constpool.extract_class pool idx
                                        , pkgs = [] }


fun num2num {start,line} =
    { start = Label.fromInt (Word16.toInt start)
    , line  = Word16.toInt line} 

fun class2class pool (info:inner_class_info) =
    { inner_info  = optional_function (idx2jclass pool)
                                      (#inner_info info)
    , outer_info  = optional_function (idx2jclass pool)
			              (#outer_info info)
    , inner_name  = optional_function (extract_name pool)
    			              (#inner_name info)
    , inner_flags = f2d_modifiers (Word16.toInt (#inner_flags info))
			                         "class" }

fun var2var pool (info:local_var_info) =
let val start      = #start info
    val start_val  = Word16.toInt start 
    val length     = #length info
    val length_val = Word16.toInt length 
    val stop       = start_val + length_val
in  { from   = w16_to_label start
    , thru   = Label.fromInt stop
    , name   = extract_name pool (#name info)
    , ty     = Parser.parseType (extract_name pool (#desc info))
    , index  = Localvar.fromInt (Word16.toInt (#index info)) }
end

fun hdls2hdls pool (hdl:exn_hdl) = { start = w16_to_label (#start hdl)
                                   , stop  = w16_to_label (#stop hdl)
		                   , entry = w16_to_label (#entry hdl)
			           , catch = optional_function (idx2jclass pool)
			                                       (#catch hdl) }

fun fattr2dattr pool attr = case attr of
                              SRCFILE  a => Classdecl.SRCFILE (extract_name pool (#file a))
			    | CONSTVAL a => Classdecl.CONSTVAL (idx2jvm_const pool (#value a))
			    | CODE     a => Classdecl.CODE { stack  = Word16.toInt (#stack a)
			                                   , locals = Word16.toInt (#locals a)
							   , code   = Parser.parseByteCode (#code a) pool
							   , hdls   = map (hdls2hdls pool) (#hdls a)
							   , attrs  = map (fattr2dattr pool) (#attrs a) } 
			    | EXNS     a => Classdecl.EXNS     (map (idx2jclass pool)  (#exns a))
			    | LINENUM  a => Classdecl.LINENUM  (map  num2num           (#lines a))
			    | LOCALVAR a => Classdecl.LOCALVAR (map (var2var pool)     (#vars a))
			    | INNER    a => Classdecl.INNER    (map (class2class pool) (#classes a))
			    | ATTR     a => Classdecl.ATTR { attr = extract_name pool (#attr a)
			                                   , info = #info a }

fun field2field pool (mem:member) = { flags = (f2d_modifiers o Word16.toInt) (#flags mem) "field"
                                    , name  = Constpool.extract_name pool (#name mem)
                                    , ty    = Parser.parseType (extract_name pool (#desc mem))
                                    , attrs = map (fattr2dattr pool) (#attrs mem) } 

fun method2method pool (mem:member) = { flags = (f2d_modifiers o Word16.toInt) (#flags mem) "method"
                                      , name  = Constpool.extract_name pool (#name mem)
                                      , msig  = Parser.parseTypes (extract_name pool (#desc mem))
                                      , attrs = map (fattr2dattr pool) (#attrs mem) }

local open Classdecl
in fun file2decl (cf:MyClassFile.class_file) =
   let val pool       = #pool cf
       val mods       = Word16.toInt (#flags cf)
       val decl_flags = f2d_modifiers mods "class"
       val decl_this  = Jvmtype.class { name = Constpool.extract_class pool (#this cf)
                                      , pkgs = [] }
       val decl_super = case (#super cf) of
                          NONE     => NONE
                        | SOME idx => SOME (idx2jclass pool idx)
       val decl_ifcs  = map (idx2jclass pool) (#ifcs cf)
       val fdecls     = map (field2field pool) (#fields cf)
       val mdecls     = map (method2method pool) (#methods cf)

       val decl_attrs = map (fattr2dattr pool) (#attrs cf)
   in  { flags  = decl_flags
       , this   = decl_this
       , super  = decl_super
       , ifcs   = decl_ifcs
       , fdecls = fdecls
       , mdecls = mdecls
       , attrs  = decl_attrs }
   end
end






(*********************************************************************************************)
(************************************ file2run functions *************************************)
(*********************************************************************************************)

val cp2rcp = (RuntimePool.makeIndex o indexValue)



fun num2run {start,line} =
    { start = Label.fromInt (Word16.toInt start)
    , line  = Word16.toInt line} 

fun class2run (info:inner_class_info) =
    { inner_info  = ((optional_function cp2rcp) o #inner_info) info
    , outer_info  = ((optional_function cp2rcp) o #outer_info) info
    , inner_name  = ((optional_function cp2rcp) o #inner_name) info
    , inner_flags = f2d_modifiers (Word16.toInt (#inner_flags info))
			                         "class" }

fun var2run (info:local_var_info) =
let val start      = #start info
    val start_val  = Word16.toInt start 
    val length     = #length info
    val length_val = Word16.toInt length 
    val stop       = start_val + length_val
in  { start  = w16_to_label start
    , length = Label.fromInt length_val
    , name   = (cp2rcp o #name) info
    , desc   = (cp2rcp o #desc) info
    , index  = Localvar.fromInt (Word16.toInt (#index info)) }
end

fun hdls2run (hdl:exn_hdl) = { start = w16_to_label (#start hdl)
                                  , stop  = w16_to_label (#stop hdl)
		                  , entry = w16_to_label (#entry hdl)
			          , catch = ((optional_function cp2rcp) o #catch) hdl }


local open Array
in fun fattr2run aref attr =
       case attr of
         SRCFILE  a => update ( aref
                              , RuntimeClass.srcfile
	 		      , RuntimeClass.SRCFILE ( (cp2rcp o #file) a ) ) 
       | CONSTVAL a => update ( aref
			      , RuntimeClass.constval
			      , RuntimeClass.CONSTVAL ( (cp2rcp o #value) a ) )
       | CODE     a => let val attr_array = Array.array (9,RuntimeClass.EMPTY)
			   val _ = List.app (fattr2run attr_array)
					    (#attrs a) 
		       in update ( aref
				 , RuntimeClass.code
				 , RuntimeClass.CODE { stack  = Word16.toInt (#stack a)
			                             , locals = Word16.toInt (#locals a)
			 		 	     , code   = Parser.parseByteCode2 (#code a)
			 			     , hdls   = map hdls2run
						                    (#hdls a)
						     , attrs  = attr_array  } )
		       end
       | EXNS     a => update ( aref
			      , RuntimeClass.exns
			      , RuntimeClass.EXNS (map cp2rcp (#exns a)) )
       | LINENUM  a => update ( aref
			      , RuntimeClass.linenum
			      , RuntimeClass.LINENUM (map num2run (#lines a)) )
       | LOCALVAR a => update ( aref
			      , RuntimeClass.localvar 
			      , RuntimeClass.LOCALVAR (map var2run (#vars a)) )
       | INNER    a => update ( aref
			      , RuntimeClass.inner
			      , RuntimeClass.INNER (map class2run (#classes a)) )
       | ATTR     a => update ( aref
			      , RuntimeClass.attr
			      , RuntimeClass.ATTR { attr = (cp2rcp o #attr) a
			                          , info = #info a } )
end

fun field2run (mem:member) = 
    let val fld_attrs = Array.array (9, RuntimeClass.EMPTY)
        val _         = app (fattr2run fld_attrs) (#attrs mem)
	val fls       = (f2d_modifiers o Word16.toInt) (#flags mem) "field"
	val idx       = (cp2rcp o #name) mem
    in  SOME { flags = fls
             , name  = idx
             , desc  = (cp2rcp o #desc) mem
             , attrs = fld_attrs }
    end

fun method2run (mem:member) =
    let val met_attrs = Array.array (9, RuntimeClass.EMPTY)
        val _         = app (fattr2run met_attrs) (#attrs mem)
	val fls       = (f2d_modifiers o Word16.toInt) (#flags mem) "method"
	val idx       = (cp2rcp o #name) mem
    in  SOME { flags = fls
             , name  = idx
             , desc  = (cp2rcp o #desc) mem
             , attrs =  met_attrs } 
    end

val exnFind = Error "HashFind trouble ..."

local 
   open RuntimeClass
   fun idxEq (RuntimePool.IDX n) (RuntimePool.IDX m) = (m = n)
   fun rcp2cp (RuntimePool.IDX i) = Constpool.makeIndex i
   fun getNameIdx pool nametype_idx =
       case Constpool.lookup pool nametype_idx of
            CPnametype {name, desc} => name
          | _                       => raise Error "CPnametype expected in Transform.sml"                   
   fun getTypeIdx pool nametype_idx =
       case Constpool.lookup pool nametype_idx of
            CPnametype {name, desc} => desc
          | _                       => raise Error "CPnametype expected in Transform.sml"
   fun getNameTypeString pool nametype_idx =
       let val desc_idx = getTypeIdx pool nametype_idx
           val name_idx = getNameIdx pool nametype_idx
	   val name = Constpool.extract_name pool name_idx
	   val desc = Constpool.extract_name pool desc_idx
       in  name ^ desc
       end 
   fun findIndex (array,i) pool
                           name_idx = let val x = ((#name o valOf) (Array.sub(array,i)))
                                      in  if idxEq x (cp2rcp name_idx) 
                                          then i
				          else if i + 1 = Array.length array
					       then i + 1
					       else findIndex (array,i+1) pool name_idx
				      end
   fun makeRCP (this,pool,m_array,f_array,m_names,f_names)
               (CPmethodref {class,nameType}) =
	          let val name_idx = getNameIdx pool nameType
		      val name = getNameTypeString pool nameType
                      val b1 = idxEq (cp2rcp class) (cp2rcp this)
		      val b2 = List.exists (fn x => name = x) m_names
                  in  if b1 andalso b2
		      then RuntimePool.RCP (RuntimePool.RCPmethodref (findIndex (m_array,0) pool name_idx))
		      else RuntimePool.CP (CPmethodref {class = class,nameType = nameType})
		  end
     | makeRCP (this,pool,m_array,f_array,m_names,f_names)
               (CPfieldref  {class,nameType}) =
	          let val name_idx = getNameIdx pool nameType
		      val name = getNameTypeString pool nameType
                      val b1 = idxEq (cp2rcp class) (cp2rcp this)
		      val b2 = List.exists (fn x => name = x) f_names
                  in  if b1 andalso b2
		      then RuntimePool.RCP (RuntimePool.RCPfieldref { fieldIndex = (findIndex (f_array,0) pool name_idx)
		                                                    , index = cp2rcp nameType })
		      else RuntimePool.CP (CPfieldref {class = class,nameType = nameType})
		  end
     | makeRCP _ x               = RuntimePool.CP x
   fun auxMakeHash pool (mem,c) hash  = let val name_string = (Constpool.extract_name pool o rcp2cp o #name o valOf) mem
                                            val type_string = (Constpool.extract_name pool o rcp2cp o #desc o valOf) mem
                                        in Polyhash.insert hash (name_string ^ type_string, c) 
					end
   fun makeHash pool (array,c) hash   = if c < Array.length array
                                        then ( auxMakeHash pool (Array.sub (array,c),c) hash ;
					       makeHash pool (array, c+1) hash )
				        else ()
in fun file2run (cf:MyClassFile.class_file) =
   let val pool      = #pool cf
       val cpl       = cp_size pool

       val newpool   = Array.array (cpl,RuntimePool.CP Constpool.CPunused)
       val poolref   = ref newpool
       val count     = ref 1
       
       val f_array   = Array.fromList(map field2run (#fields cf))
       val f_hash    = Polyhash.mkPolyTable(Array.length f_array, findExn)
       val _         = makeHash pool (f_array,0) f_hash
       val f_count   = ref 0

       val m_array   = Array.fromList (map method2run (#methods cf))
       val m_hash    = Polyhash.mkPolyTable(Array.length m_array, findExn)
       val _         = makeHash pool (m_array,0) m_hash
       val m_count   = ref 0

       fun m2n m = (extract_name pool (#name m)) ^ (extract_name pool (#desc m))
      
       val m_name_list = map m2n (#methods cf)
       val f_name_list = map m2n (#fields cf)

       val _  = while (!count < cpl) do
                ( Array.update ( newpool
		               , !count
			       , makeRCP ( #this cf
			                 , pool
			                 , m_array
					 , f_array
					 , m_name_list
					 , f_name_list )
					 (lookup pool (makeIndex (!count))) );
		 count := !count + 1 )

       val mods      = Word16.toInt (#flags cf)
       val run_flags = f2d_modifiers mods "class"
       
       
       val aref      = (Array.array (9, EMPTY))
       val _         = app (fattr2run aref) (#attrs cf)
   in  { class_ref = 0 
       , pool    = !poolref
       , flags   = run_flags
       , this    = (cp2rcp o #this) cf
       , super   = ((optional_function cp2rcp) o #super) cf
       , ifcs    = map cp2rcp (#ifcs cf)
       , fields  = { array = f_array
                   , hash  = f_hash }
       , methods = { array = m_array
                   , hash  = m_hash }
       , attrs   = aref }
   end
end




end
